import numpy as np
from ..visera.Quantization import fixedBinSizeQuantization, uniformQuantization, lloydQuantization
from ..visera.interpolation import imresize3D, imresize
import logging
import gc
from ...utils.utils import save_numpy_on_disk, remove_temp_file
# -------------------------------------------------------------------------
# function [ROIonly,levels,Maskout,ROIboxResmp,newpixelW,newsliceTh] =
# prepareVolume(volume, Mask, DataType, pixelW, sliceTh, newVoxelSize,
#               VoxInterp, ROIInterp, ROI_PV, scaleType, isIsot2D, isScale,
#               isGLround, DiscType, qntz,Bin, isReSegRng, ReSegIntrvl,
#               isOutliers)
# -------------------------------------------------------------------------
# DESCRIPTION:
# This function prepares the input volume for 3D texture analysis. The
# following operations are performed:
#
# 1. Pre-processing of the ROIbox (PET: square-root (default:off), MR:
#    Collewet normalizaton, CT: nothing).
# 2. Isotropic resampling.
# 3. Image resegmentation.
# 4. grey levels rounding.
# 5. Image resegmentation.
# 6. Quantization of intensity dynamic range.
# 7. Crop the ROI to get rid of extra zeros
# --> This function is compatible with both 2D and 3D analysis
# -------------------------------------------------------------------------
# INPUTS:
# - volume: 2D or 3D array containing the medical images to analyze
# - mask: 2D or 3D array of dimensions corresponding to 'volume'. The mask
#         contains 1's in the region of interest (ROI), and 0's elsewhere.
# - DataType: String specifying the type of scan analyzed. Either 'PET',
#             'CT' or 'MRscan'.
# - pixelW: Numerical value specifying the in-plane resolution (mm) of 'volume'.
# - sliceTh: Numerical value specifying the slice spacing (mm) of 'volume'.
#           Put a random number for 2D analysis.
# - newVoxelSize: Numerical value specifying the scale at which 'volume' is
#                 isotropically  resampled in 3D (mm).
# - VoxInterp: interpolation method for the intensity ROI
# - ROIInterp: interpolation method for the morphological ROI
# - ROI_PV: partial volume threshold for thresholding morphological ROI.
#           Used to threshold ROI after resampling:
#           i.e. ROI(ROI<ROI_PV) =0, ROI(ROI>ROI_PV) = 1.
# - scaleType: 'NoRescale' if no resampling, 'XYZscale' if resample in all
#              3 dimensions, 'XYscale' if only resampling in X and Y
#              dimensions, 'Zscale' if only scaling Z direction.
# - isIsot2D: =1 if resampling only in X and Y dimensions.
# - isScale: whether to perform resampling
# - isGLrounding: whether to perform grey level rounding
# - DiscType: discritization type. Either 'FNB' for fixed number of bins
#             or 'FBS' for fixed bin size.
# - qntz: String specifying the quantization algorithm to use on
#             'volume'. Either 'Lloyd' for Lloyd-Max quantization, or
#             'Uniform' for uniform quantization.
# - Bin: number of bins (for fixed number of bins method) or bin size
#            (for bin size method).
# - isReSeg: whether to perform resegmentation
# - ResegIntrval: a 1x2 array of values expressing resegmentation interval.
# - isOutliers: whether to perform instensity outlier filtering
# -------------------------------------------------------------------------
# OUTPUTS: ROIonly,levels,Maskout,ROIboxResmp,newpixelW,newsliceTh
# - ImgBoxResampQuntz3D: Smallest box containing the ROI, with the imaging
#           data of the ready for texture analysis computations. Voxels
#           outside the ROI are set to NaNs.
# - levels: Vector containing the quantized gray-levels in the tumor region
#           (or reconstruction levels of quantization).
# - Maskout: Smallest matrix containing morphological ROI (0 or 1)
# - ROIboxResmp: Smallest matrix containing intensity ROI. This is the
#                resampled ROI, went through resegmtation and GL rounding,
#                and just before quantization.
# - newpixelW: width of the voxel in the X (=Y) direction in resampled ROI
# - newsliceTh: Slice thickness of the voxel in the Z direction in resamped
#               ROI
# -------------------------------------------------------------------------
# REFERENCE:
# [1] Vallieres, M. et al. (2015). A radiomics model from joint FDG-PET and
#     MRI texture features for the prediction of lung metastases in soft-tissue
#     sarcomas of the extremities. Physics in Medicine and Biology, 60(14),
#     5471-5496. doi:10.1088/0031-9155/60/14/5471
# -------------------------------------------------------------------------
# AUTHOR(S):
# - Martin Vallieres <mart.vallieres@gmail.com>
# - Saeed Ashrafinia
# - Mahdi Hosseinzadeh
# -------------------------------------------------------------------------
# HISTORY:
# - Creation: December 2022
# -------------------------------------------------------------------------


def prepareVolume(volume, mask_input, DataType, pixelW, sliceTh,
                  newVoxelSize, VoxInterp, ROIInterp, ROI_PV, scaleType, isIsot2D,
                  isScale, isGLround, DiscType, qntz, Bin,
                  isReSegRng, ReSegIntrvl, isOutliers):
    """Memory-optimized volume preparation with reduced copying and efficient operations"""

    # Set up quantization function
    if DiscType == 'FBS':
        quantization = fixedBinSizeQuantization
    elif DiscType == 'FBN':
        quantization = uniformQuantization
    else:
        raise('Error with discretization type. Must either be "FBS" (Fixed Bin Size) or "FBN" (Fixed Number of Bins).')

    if qntz == 'Lloyd':
        quantization = lloydQuantization

    # Use views instead of copies where possible
    # ROIBox = Mask if Mask.flags.writeable else Mask.copy()        # toto CMed

    ROIBox_path = mask_input
    # Load mask
    Mask = np.load(mask_input, mmap_mode='r')
    ROIBox = np.load(ROIBox_path, mmap_mode='r')

    Imgbox = volume.astype(np.float32) if volume.dtype != np.float32 else volume

    # MR scan specific processing
    if DataType == 'MRscan':
        ROIonly = Imgbox.copy()
        ROIonly[ROIBox == 0] = np.nan
        temp = CollewetNorm(ROIonly)
        ROIBox[np.isnan(temp)] = 0
        del temp, ROIonly
        gc.collect()

    flagPW = 0
    if scaleType == 'NoRescale':
        flagPW = 0
    elif scaleType == 'XYZscale':
        flagPW = 1
    elif scaleType == 'XYscale':
        flagPW = 2
    elif scaleType == 'Zscale':
        flagPW = 3

    if isIsot2D == 1:
        flagPW = 2

    if isScale == 0:
        flagPW = 0

    if flagPW == 0:
        a = 1
        b = 1
        c = 1
    elif flagPW == 1:
        a = pixelW / newVoxelSize
        b = pixelW / newVoxelSize
        c = sliceTh / newVoxelSize
    elif flagPW == 2:
        a = pixelW / newVoxelSize
        b = pixelW / newVoxelSize
        c = 1
    elif flagPW == 3:
        a = 1
        b = 1
        c = sliceTh / pixelW

    # Resampling
    ImgBoxResmp = Imgbox.copy()
    ImgWholeResmp = volume.copy()
    ROIBoxResmp = ROIBox.copy()
    ROIwholeResmp = Mask.copy()

    if Imgbox.ndim == 3 and flagPW != 0:
        if (a + b + c) != 3:
            ROIBoxResmp_path = imresize3D(ROIBox_path, [pixelW, pixelW, sliceTh],
                                     [np.ceil(ROIBox.shape[0] * a), np.ceil(ROIBox.shape[1] * b),
                                      np.ceil(ROIBox.shape[2] * c)], ROIInterp, 'constant',
                                     [newVoxelSize, newVoxelSize, newVoxelSize], isIsot2D, use_disk=True)
            ROIBoxResmp = np.load(ROIBoxResmp_path, mmap_mode='r+')
            Imgbox[np.isnan(Imgbox)] = 0
            ImgBoxResmp = imresize3D(Imgbox, [pixelW, pixelW, sliceTh],
                                     [np.ceil(Imgbox.shape[0] * a), np.ceil(Imgbox.shape[1] * b),
                                      np.ceil(Imgbox.shape[2] * c)], VoxInterp, 'constant',
                                     [newVoxelSize, newVoxelSize, newVoxelSize], isIsot2D, use_disk=False)
            ROIBoxResmp[ROIBoxResmp < ROI_PV] = 0
            ROIBoxResmp[ROIBoxResmp >= ROI_PV] = 1
            ROIwholeResmp_path = imresize3D(mask_input, [pixelW, pixelW, sliceTh],
                                       [np.ceil(Mask.shape[0] * a), np.ceil(Mask.shape[1] * b),
                                        np.ceil(Mask.shape[2] * c)], ROIInterp, 'constant',
                                       [newVoxelSize, newVoxelSize, newVoxelSize], isIsot2D, use_disk=True)
            ROIwholeResmp = np.load(ROIwholeResmp_path, mmap_mode='r+')
            ImgWholeResmp = imresize3D(volume, [pixelW, pixelW, sliceTh],
                                       [np.ceil(volume.shape[0] * a), np.ceil(volume.shape[1] * b),
                                        np.ceil(volume.shape[2] * c)], VoxInterp, 'constant',
                                       [newVoxelSize, newVoxelSize, newVoxelSize], isIsot2D, use_disk=False)
            if np.max(ROIwholeResmp) < ROI_PV:
                import logging
                logging.getLogger("Dev_logger").warning("PREPARE_VOLUME- Resampled ROI has no voxels with value above ROI_PV. Cutting ROI_PV to half.")
                ROI_PV = ROI_PV / 2

    
            ROIwholeResmp[ROIwholeResmp < ROI_PV] = 0
            ROIwholeResmp[ROIwholeResmp >= ROI_PV] = 1

    elif Imgbox.ndim == 2 and flagPW != 0:
        if (a + b) != 2:
            ROIBoxResmp_path = imresize(ROIBox_path, [pixelW, pixelW],
                                   [np.ceil(ROIBox.shape[0] * a), np.ceil(ROIBox.shape[1] * b)], ROIInterp,
                                   [newVoxelSize, newVoxelSize], use_disk=True)
            ROIBoxResmp = np.load(ROIBoxResmp_path, mmap_mode='r+')
            ImgBoxResmp = imresize(Imgbox, [pixelW, pixelW],
                                   [np.ceil(Imgbox.shape[0] * a), np.ceil(Imgbox.shape[1] * b)], VoxInterp,
                                   [newVoxelSize, newVoxelSize], use_disk=False)
            ROIBoxResmp[ROIBoxResmp < ROI_PV] = 0
            ROIBoxResmp[ROIBoxResmp >= ROI_PV] = 1
            
            ROIwholeResmp_path = imresize(mask_input, [pixelW, pixelW], [np.ceil(Mask.shape[0] * a), np.ceil(Mask.shape[1] * b)],
                                     ROIInterp, [newVoxelSize, newVoxelSize], use_disk=True)
            ROIwholeResmp = np.load(ROIwholeResmp_path, mmap_mode='r+')
            ImgWholeResmp = imresize(volume, [pixelW, pixelW],
                                     [np.ceil(volume.shape[0] * a), np.ceil(volume.shape[1] * b)], VoxInterp,
                                     [newVoxelSize, newVoxelSize], use_disk=False)
            if np.max(ROIwholeResmp) < ROI_PV:
                import logging
                logging.getLogger("Dev_logger").warning("PREPARE_VOLUME- Resampled ROI has no voxels with value above ROI_PV. Cutting ROI_PV to half.")
                ROI_PV = ROI_PV / 2

            ROIwholeResmp[ROIwholeResmp < ROI_PV] = 0
            ROIwholeResmp[ROIwholeResmp >= ROI_PV] = 1
    else:
        ROIBoxResmp_path = None
        ROIwholeResmp_path = None

    IntsBoxROI = ImgBoxResmp.copy()

    ImgBoxResmp[ROIBoxResmp == 0] = np.nan

    IntsBoxROI = roundGL(ImgBoxResmp, isGLround)
    ImgWholeResmp = roundGL(ImgWholeResmp, isGLround)

    IntsBoxROItmp1 = IntsBoxROI.copy()
    ImgWholeResmptmp1 = ImgWholeResmp.copy()
    IntsBoxROItmp2 = IntsBoxROI.copy()
    ImgWholeResmptmp2 = ImgWholeResmp.copy()

    if isReSegRng == 1:
        IntsBoxROItmp1[IntsBoxROI < ReSegIntrvl[0]] = np.nan
        IntsBoxROItmp1[IntsBoxROI > ReSegIntrvl[1]] = np.nan
        ImgWholeResmptmp1[ImgWholeResmp < ReSegIntrvl[0]] = np.nan
        ImgWholeResmptmp1[ImgWholeResmp > ReSegIntrvl[1]] = np.nan

    if isOutliers == 1:
        Mu = np.nanmean(IntsBoxROI)
        Sigma = np.nanstd(IntsBoxROI)
        IntsBoxROItmp2[IntsBoxROI < (Mu - 3 * Sigma)] = np.nan
        IntsBoxROItmp2[IntsBoxROI > (Mu + 3 * Sigma)] = np.nan

        Mu = np.nanmean(ImgWholeResmp)
        Sigma = np.nanstd(ImgWholeResmp)
        ImgWholeResmptmp2[ImgWholeResmp < (Mu - 3 * Sigma)] = np.nan
        ImgWholeResmptmp2[ImgWholeResmp > (Mu + 3 * Sigma)] = np.nan

    IntsBoxROI = getMutualROI(IntsBoxROItmp1, IntsBoxROItmp2)
    ImgWholeResmp = getMutualROI(ImgWholeResmptmp1, ImgWholeResmptmp2)

    newpixelW = pixelW / a
    newsliceTh = sliceTh / c

    # Determine minimum GL value efficiently
    if DataType == 'PET':
        minGL = 0
    elif DataType == 'CT':
        minGL = ReSegIntrvl[0] if isReSegRng == 1 else np.nanmin(IntsBoxROI)
    else:
        minGL = np.nanmin(IntsBoxROI)

    # Perform quantization
    ImgBoxResampQuntz3D, levels = quantization(IntsBoxROI, Bin, minGL)

    # Calculate bounding box and crop efficiently

    boxBound = computeBoundingBox(ROIBoxResmp)
    MorphROI = ROIBoxResmp[boxBound[0][0]:boxBound[0][1], boxBound[1][0]:boxBound[1][1], boxBound[2][0]:boxBound[2][1]]
    IntsBoxROI = IntsBoxROI[boxBound[0][0]:boxBound[0][1], boxBound[1][0]:boxBound[1][1], boxBound[2][0]:boxBound[2][1]]
    ImgBoxResampQuntz3D = ImgBoxResampQuntz3D[boxBound[0][0]:boxBound[0][1], boxBound[1][0]:boxBound[1][1],
                          boxBound[2][0]:boxBound[2][1]]


    # Save numpy arrays on disk
    ROIwholeResmp_path2 = save_numpy_on_disk(ROIwholeResmp, prefix='final_ROIwholeResmp', suffix='.npy')
    MorphROI_path = save_numpy_on_disk(MorphROI, prefix='final_MorphROI', suffix='.npy')
    # Clean RAM
    del ROIBoxResmp, ROIwholeResmp, MorphROI, ROIBox, Mask
    gc.collect()
    # Clean disk
    if ROIBoxResmp_path and ROIwholeResmp_path:
        remove_temp_file(ROIBoxResmp_path)
        remove_temp_file(ROIwholeResmp_path)

    return ImgBoxResampQuntz3D, levels, MorphROI_path, IntsBoxROI, ImgWholeResmp, ROIwholeResmp_path2, newpixelW, newsliceTh


# -------------------------------------------------------------------------
# function [boxBound] = computeBoundingBox(mask)
# -------------------------------------------------------------------------
# DESCRIPTION:
# This function computes the smallest box containing the whole region of
# interest (ROI).
# -------------------------------------------------------------------------
# INPUTS:
# - mask: 3D array, with 1's inside the ROI, and 0's outside the ROI.
# -------------------------------------------------------------------------
# OUTPUTS:
# - boxBound: Bounds of the smallest box containing the ROI.
#             Format: [minRow, maxRow;
#                      minColumn, maxColumns;
#                      minSlice, maxSlice]
# -------------------------------------------------------------------------
# # AUTHOR(S):
# - Mahdi Hosseinzadeh
# -------------------------------------------------------------------------
# HISTORY:
# - Creation: December 2022
# --------------------------------------------------------------------------

def computeBoundingBox(Data_ROI_mat):
    # Load roi from disk        toto
    # Data_ROI_mat = np.load(Data_ROI_mat_path, mmap_mode='r')
    # Efficient and safe computation of bounding box for non-zero elements
    indices = np.where(Data_ROI_mat != 0)
    if len(indices[0]) == 0:
        return np.zeros((3, 2), dtype=np.uint32)

    boxBound = np.empty((3, 2), dtype=np.uint32)
    for dim in range(3):
        boxBound[dim, 0] = np.min(indices[dim])
        boxBound[dim, 1] = np.max(indices[dim]) + 1
    # # Clean RAM
    # del Data_ROI_mat
    # gc.collect()
    return boxBound


def CollewetNorm(ROIonly):
    temp = ROIonly[~np.isnan(ROIonly)]
    u = np.mean(temp)
    sigma = np.std(temp)

    ROIonlyNorm = ROIonly.copy()
    ROIonlyNorm[ROIonly > (u + 3 * sigma)] = np.nan
    ROIonlyNorm[ROIonly < (u - 3 * sigma)] = np.nan

    return ROIonlyNorm


# This function rounds image intensity voxels to the nearest integer.
def roundGL(Img, isGLrounding):
    """Memory-efficient rounding with in-place operations when possible"""
    if isGLrounding == 1:
        GLroundedImg = np.round(Img)
    else:
        GLroundedImg = Img.copy()

    return GLroundedImg


# This function receives two ROIs as a result of two resegmentation methods
# (range reseg. and outlier reseg.) and return a third ROI containing the
# mutual voxels inside both ROIs.
def getMutualROI(ROI1, ROI2):
    tmp1 = np.multiply(ROI1, ROI2)
    tmp2 = tmp1.copy()
    tmp2[~np.isnan(tmp1)] = 1
    outROI = np.multiply(tmp2, ROI1)

    return outROI


def getImgBox(volume, mask_input, isReSeg, ResegIntrval):
    # Load mask from disk
    mask = np.load(mask_input, mmap_mode='r')

    boxBound = computeBoundingBox(mask)
    # maskBox = mask[boxBound[0][0]:boxBound[0][1],boxBound[1][0]:boxBound[1][1],boxBound[2][0]:boxBound[2][1]]
    SUVbox = volume[boxBound[0][0]:boxBound[0][1], boxBound[1][0]:boxBound[1][1], boxBound[2][0]:boxBound[2][1]]

    if isReSeg == 1:
        SUVbox[SUVbox < ResegIntrval[0]] = np.nan
        SUVbox[SUVbox > ResegIntrval[1]] = np.nan

    # Clean RAM
    del mask
    gc.collect()

    return SUVbox



# -------------------------------------------------------------------------
# function [boxBound] = computeBoundingBox(mask)
# -------------------------------------------------------------------------
# DESCRIPTION:
# This function computes the smallest box containing the whole region of
# interest (ROI).
# -------------------------------------------------------------------------
# INPUTS:
# - mask: 3D array, with 1's inside the ROI, and 0's outside the ROI.
# -------------------------------------------------------------------------
# OUTPUTS:
# - boxBound: Bounds of the smallest box containing the ROI.
#             Format: [minRow, maxRow;
#                      minColumn, maxColumns;
#                      minSlice, maxSlice]
# -------------------------------------------------------------------------
# # AUTHOR(S):
# - Mahdi Hosseinzadeh
# -------------------------------------------------------------------------
# HISTORY:
# - Creation: December 2022
# --------------------------------------------------------------------------


def CollewetNorm(ROIonly):
    temp = ROIonly[~np.isnan(ROIonly)]
    u = np.mean(temp)
    sigma = np.std(temp)

    ROIonlyNorm = ROIonly.copy()
    ROIonlyNorm[ROIonly > (u + 3 * sigma)] = np.nan
    ROIonlyNorm[ROIonly < (u - 3 * sigma)] = np.nan

    return ROIonlyNorm






from scipy.sparse import spdiags
import numpy as np


# -------------------------------------------------------------------------
# function [GLRLM2D_Cmb, GLRLM2D_Avg] = getGLRLM3Dtex(ROIonly,levels)
# -------------------------------------------------------------------------
# DESCRIPTION: 
# This function calculates the GLRLM matrix for each slice of an ROI.
# Every slice is calculated separately, then features are calculated.
# 
# Like the grey level co-occurrence matrix, GLRLM also assesses the
# distribution of discretised grey levels in an image or in a stack of
# images. However, instead of assessing the combination of levels between
# neighbouring pixels or voxels, GLRLM assesses grey level run # lengths.
# Run length counts the frequency of consecutive voxels with discretised
# grey level i along  direction \Delta.
# -------------------------------------------------------------------------
# INPUTS:
# - ROIonly: Smallest box containing the ROI, with the imaging data ready 
#            for texture analysis computations. Voxels outside the ROI are 
#            set to NaNs. 
# - levels: number of bins (for fixed number of bins method) or bin size
#           (for bin size method) 
# Note: ROIonly is the outputs of prepareVolume.m
# -------------------------------------------------------------------------
# OUTPUTS:
# - GLRLM3D_Cmb: 3D GLRLM features: First merging GLRLMs for all directions
#               then calculate features for the combined GLRLM matrix.
# - GLRLM3D_Avg: 3D GLRLM features calculate GLRLM features for each
#               direction, then average over all directions.
# -------------------------------------------------------------------------
# AUTHOR(S): 
# - Saeed Ashrafinia
# - Mahdi Hosseinzadeh
# -------------------------------------------------------------------------
# HISTORY:
# - Creation: December 2022
# -------------------------------------------------------------------------


def getGLRLM3Dtex(ROIonly2,levels2, feature_value_mode='REAL_VALUE'):

    levels = levels2.copy()

    ROIonly = ROIonly2.copy()
    GLRLM, GLRLM_D = getGLRLM(ROIonly , levels)


    GLRLM2D_Cmb = CalcGLRLM(GLRLM, ROIonly, feature_value_mode=feature_value_mode)
    FeatsTmp = CalcGLRLM(GLRLM_D, ROIonly, feature_value_mode=feature_value_mode)
    # FeatsTmp = FeatsTmp[FeatsTmp != np.nan]
    GLRLM2D_Avg = np.nanmean(FeatsTmp , axis=1)

    return GLRLM2D_Cmb, GLRLM2D_Avg


# -------------------------------------------------------------------------
# function [GLRLM2D_Cmb, GLRLM2D_Avg] = getGLRLM2Dtex(ROIonly,levels)
# -------------------------------------------------------------------------
# DESCRIPTION: 
# This function calculates the GLRLM matrix for each slice of an ROI.
# Every slice is calculated separately, then features are calculated.
# 
# Like the grey level co-occurrence matrix, GLRLM also assesses the
# distribution of discretised grey levels in an image or in a stack of
# images. However, instead of assessing the combination of levels between
# neighbouring pixels or voxels, GLRLM assesses grey level run # lengths.
# Run length counts the frequency of consecutive voxels with discretised
# grey level i along  direction \Delta.
# -------------------------------------------------------------------------
# INPUTS:
# - ROIonly: Smallest box containing the ROI, with the imaging data ready 
#            for texture analysis computations. Voxels outside the ROI are 
#            set to NaNs. 
# - levels: number of bins (for fixed number of bins method) or bin size
#           (for bin size method) 
# Note: RIOonly is the outputs of prepareVolume.m
# -------------------------------------------------------------------------
# OUTPUTS:
# - GLRLM2D_Cmb: 2D GLRLM features: merging GLRLM for different slice of
#                 the volume, then calculate features for the combined
#                 GLRLM matrix. 
# - GLRLM2D_Avg: 2D GLRLM features: calculate GLRLM features for each
#                slice first, then average over all slices.
# -------------------------------------------------------------------------
# AUTHOR(S): 
# - Saeed Ashrafinia
# - Mahdi Hosseinzadeh
# -------------------------------------------------------------------------
# HISTORY:
# - Creation: December 2022
# -------------------------------------------------------------------------


def getGLRLM2Dtex(ROIonly2,levels2, feature_value_mode='REAL_VALUE'):

    levels = levels2.copy()
    ROIonly = ROIonly2.copy()
    nx, ny, nZ = ROIonly.shape
    FeatTmp = []
    FeatTmp_D = []
    GLRLM_all = np.zeros((levels.shape[0] , np.max([nx,ny]) , nZ), dtype=np.float32)
    GLRLM_D_all = np.zeros((levels.shape[0] , np.max([nx,ny]) , 4 , nZ), dtype=np.float32)
    for s in range(0,nZ):
        GLRLM, GLRLM_D = getGLRLM(ROIonly[:,:,s] , levels)
        GLRLMCal =  CalcGLRLM(GLRLM, ROIonly[:,:,s], feature_value_mode=feature_value_mode)
        GLRLM_D_Cal = CalcGLRLM(GLRLM_D[:,:,0:4], ROIonly[:,:,s], feature_value_mode=feature_value_mode)

        if s == 0:
            FeatTmp = GLRLMCal
            FeatTmp_D = GLRLM_D_Cal
        else:    
            FeatTmp = np.column_stack((FeatTmp , GLRLMCal))
            FeatTmp_D = np.dstack((FeatTmp_D , GLRLM_D_Cal))

        GLRLM_all[:,0:GLRLM.shape[1],s] = GLRLM
        GLRLM_D_all[:,0:GLRLM_D.shape[1],0:4,s] = GLRLM_D[:,:,0:4]

    if FeatTmp.ndim == 1:
        FeatTmp = FeatTmp[:, np.newaxis]
    Feats_MSKD = np.nanmean(FeatTmp , axis=1)

    try:
        Feats_MSKD = np.nanmean(Feats_MSKD , axis=1)
    except:
        Feats_MSKD = Feats_MSKD

    Feats_MSKD[12] = Feats_MSKD[12] /4

    if np.ndim(FeatTmp_D) == 2:
        FeatTmp_D = FeatTmp_D[:, :, np.newaxis]

    FeatsTmp2_D = np.reshape( a= FeatTmp_D, newshape=(16,FeatTmp_D.shape[1]*FeatTmp_D.shape[2]),order='F')
    Feats_KSKD = np.nanmean(FeatsTmp2_D,axis=1)
    try:
        Feats_KSKD = np.nanmean(Feats_KSKD , axis=1)
    except:
        Feats_KSKD = Feats_KSKD

    GLRLM_MergeSlice_KeepDirs = np.zeros((GLRLM_D_all.shape[0], GLRLM_D_all.shape[1],GLRLM_D_all.shape[2]), dtype=np.float32)
    GLRLM_MergeSlice_KeepDirs[:,:,:] = np.sum(GLRLM_D_all , axis=3)
    tmp = np.sum(np.sum(GLRLM_MergeSlice_KeepDirs , axis=0),axis=0)

    GLRLMnorm_MergeSlice_KeepDirs = np.divide(GLRLM_MergeSlice_KeepDirs , np.tile(tmp ,(GLRLM_D_all.shape[0], GLRLM_D_all.shape[1],1)))
    GLRLMnorm_MergeSlice_KeepDirs = np.nan_to_num(GLRLMnorm_MergeSlice_KeepDirs, nan=0)

    Feats_KSMD = CalcGLRLM(GLRLMnorm_MergeSlice_KeepDirs, ROIonly, feature_value_mode=feature_value_mode)
    try:
        Feats_KSMD = np.nanmean(Feats_KSMD , axis=1)
    except:
        Feats_KSMD = Feats_KSMD

    GLRLM_AllMerged   = np.sum(np.sum(GLRLM_D_all,axis=2),axis=2)
    GLRLMnorm_AllMerged = GLRLM_AllMerged / np.sum(GLRLM_AllMerged)

    Feats_MSMD = CalcGLRLM(GLRLMnorm_AllMerged, ROIonly, feature_value_mode=feature_value_mode)
    try:
        Feats_MSMD = np.nanmean(Feats_MSMD , axis=1)
    except:
        Feats_MSMD = Feats_MSMD


    return Feats_KSKD, Feats_MSKD, Feats_KSMD, Feats_MSMD



def diagk(X,k):
    if X.ndim == 2:
        D = np.diag(X,k)
    else:
        is_all_zero = np.all((X == 0))
        if is_all_zero == False and 0 <= k and 1+k <= X.shape[1]:
            D = X[1+k]
        elif is_all_zero == False and k < 0 and 1-k <= X.shape[0]:
            D = X[1-k]
        else:
            # D = np.zeros((X.shape[0],1))
            D = np.array(0)
    
    return D



def spdiags(arg1, arg2 = None):

    A = arg1
    if arg2 is None:
        i,j = np.where(A)
        d = np.sort(j-i)
        index_len = np.insert(d,0,A.shape[0])
        lenn = np.diff( index_len )
        d = d [ lenn  != 0  ]
    else:
        d = arg2

    m,n = A.shape
    p = d.shape[0]
    B = np.zeros((np.min([m,n]),p), dtype=np.float32)

    for k in range(0,p):
        if m >= n:
            i = np.arange(np.max([1,1+d[k]]),np.min([n,m+d[k]])+1,1)
        else:
            i = np.arange(np.max([1,1-d[k]]),np.min([m,n-d[k]])+1,1)
        
        i = list(i - 1)
        i = [int(item) for item in i]

        B[i,k] = diagk(A,d[k])
    
    res1 = B
    res2 = d

    return res1
    # return res1,res2

# def accum(accmap, a, func=None, size=None, fill_value=0, dtype=None):
#     """
#     An accumulation function similar to Matlab's `accumarray` function.

#     Parameters
#     ----------
#     accmap : ndarray
#         This is the "accumulation map".  It maps input (i.e. indices into
#         `a`) to their destination in the output array.  The first `a.ndim`
#         dimensions of `accmap` must be the same as `a.shape`.  That is,
#         `accmap.shape[:a.ndim]` must equal `a.shape`.  For example, if `a`
#         has shape (15,4), then `accmap.shape[:2]` must equal (15,4).  In this
#         case `accmap[i,j]` gives the index into the output array where
#         element (i,j) of `a` is to be accumulated.  If the output is, say,
#         a 2D, then `accmap` must have shape (15,4,2).  The value in the
#         last dimension give indices into the output array. If the output is
#         1D, then the shape of `accmap` can be either (15,4) or (15,4,1) 
#     a : ndarray
#         The input data to be accumulated.
#     func : callable or None
#         The accumulation function.  The function will be passed a list
#         of values from `a` to be accumulated.
#         If None, numpy.sum is assumed.
#     size : ndarray or None
#         The size of the output array.  If None, the size will be determined
#         from `accmap`.
#     fill_value : scalar
#         The default value for elements of the output array. 
#     dtype : numpy data type, or None
#         The data type of the output array.  If None, the data type of
#         `a` is used.

#     Returns
#     -------
#     out : ndarray
#         The accumulated results.

#         The shape of `out` is `size` if `size` is given.  Otherwise the
#         shape is determined by the (lexicographically) largest indices of
#         the output found in `accmap`.


#     Examples
#     --------
#     >>> from numpy import array, prod
#     >>> a = array([[1,2,3],[4,-1,6],[-1,8,9]])
#     >>> a
#     array([[ 1,  2,  3],
#            [ 4, -1,  6],
#            [-1,  8,  9]])
#     >>> # Sum the diagonals.
#     >>> accmap = array([[0,1,2],[2,0,1],[1,2,0]])
#     >>> s = accum(accmap, a)
#     array([9, 7, 15])
#     >>> # A 2D output, from sub-arrays with shapes and positions like this:
#     >>> # [ (2,2) (2,1)]
#     >>> # [ (1,2) (1,1)]
#     >>> accmap = array([
#             [[0,0],[0,0],[0,1]],
#             [[0,0],[0,0],[0,1]],
#             [[1,0],[1,0],[1,1]],
#         ])
#     >>> # Accumulate using a product.
#     >>> accum(accmap, a, func=prod, dtype=float)
#     array([[ -8.,  18.],
#            [ -8.,   9.]])
#     >>> # Same accmap, but create an array of lists of values.
#     >>> accum(accmap, a, func=lambda x: x, dtype='O')
#     array([[[1, 2, 4, -1], [3, 6]],
#            [[-1, 8], [9]]], dtype=object)
#     """

#     # Check for bad arguments and handle the defaults.
#     if accmap.shape[:a.ndim] != a.shape:
#         raise ValueError("The initial dimensions of accmap must be the same as a.shape")
#     if func is None:
#         func = np.sum
#     if dtype is None:
#         dtype = a.dtype
#     if accmap.shape == a.shape:
#         accmap = np.expand_dims(accmap, -1)
#     adims = tuple(range(a.ndim))
#     if size is None:
#         size = 1 + np.squeeze(np.apply_over_axes(np.max, accmap, axes=adims))
#     size = np.atleast_1d(size)

#     # Create an array of python lists of values.
#     vals = np.empty(size, dtype='O')
#     for s in product(*[range(k) for k in size]):
#         vals[s] = []
#     for s in product(*[range(k) for k in a.shape]):
#         indx = tuple(accmap[s])
#         val = a[s]
#         vals[indx].append(val)

#     # Create the output array.
#     out = np.empty(size, dtype=dtype)
#     for s in product(*[range(k) for k in size]):
#         if vals[s] == []:
#             out[s] = fill_value
#         else:
#             out[s] = func(vals[s])

#     return out



def accum(accmap = [],size=None):

    out = np.zeros((size[0],size[1]), dtype=np.int32)

    for i in accmap:
        out[i[0]][i[1]] = out[i[0]][i[1]] + 1 


    return out

def rle_0(si,NL):


    m,n=si.shape

    oneglrlm=np.zeros((NL,n), dtype=np.float32)

    for i in range(0,m):
        x=si[i,:]
        a = x[:-1]
        b = x[1:]
        find = np.where( a != b )[0]
        find = find + 1
        if find.shape[0] != 0:
            index = np.insert(find,-1,x.shape[0])
            index = np.sort(index)
        else:
            index = []
            index.append(x.shape[0])
            
            
        # if  isinstance(index,collections.abc.Sequence) == False: 
        # index = [0,index]
        index_len = index.copy()
        index_len = np.insert(index_len,0,0)
        lenn = np.diff( index_len )
        # else:
        #     index = np.insert(index,0,0)
        #     lenn = np.diff( index )
        
        accmap = []

        indexmiues = list(np.array(index) - 1)
        lenn = list(np.array(lenn) - 1)
        # indexmiues = np.array(index) - 1
        # indexmiues = np.array(index) - 1

        # if lenn.shape[0] == 1:    
        val = x[indexmiues] - 1
        val = [int(item) for item in val]
            # accmap.append([val ,lenn]) 
        # else:
        for gg in range(0,len(lenn)):
            accmap.append([val[gg] ,lenn[gg]])      


        temp = accum(accmap = accmap,size=[NL, n]) 
        # # temp = accum(a=x, accmap=np.transpose([val,lenn]),func = np.sum ,size =[NL, n] )
        oneglrlm = temp + oneglrlm
    

    return oneglrlm

def FindMaxLength(lst):
    # maxList = max((x) for x in lst)
    maxLength = max(len(x) for x in lst )
 
    return maxLength

def rle_45(seq,NL):

    m = len(seq)
    n = FindMaxLength(seq)

    oneglrlm= np.zeros((NL,n), dtype=np.float32)

    for i in range(0,m):
        x=seq[i]

        a = x[:-1]
        b = x[1:]
        find = np.where( a != b )[0]
        find = find + 1
        if find.shape[0] != 0:
            index = np.insert(find,-1,x.shape[0])
            index = np.sort(index)
        else:
            index = []
            index.append(x.shape[0])
            
        index_len = index.copy()
        index_len = np.insert(index_len,0,0)
        lenn = np.diff( index_len )
        
        accmap = []

        indexmiues = list(np.array(index) - 1)
        lenn = list(np.array(lenn) - 1)


        val = x[indexmiues] - 1
        val = [int(item) for item in val]

        for gg in range(0,len(lenn)):
            accmap.append([val[gg] ,lenn[gg]])      


        temp = accum(accmap = accmap,size=[NL, n]) 

        oneglrlm = temp + oneglrlm
    

    return oneglrlm


def zigzag(SI):

    seq = []

    c = 1
    r = 1

    rmin = 1
    cmin = 1

    rmax = SI.shape[0]
    cmax = SI.shape[1]

    
    i = 1
    j = 1

    sq_up_begin=1

    sq_down_begin=1


    output = np.zeros((1, rmax * cmax), dtype=np.float32).flatten(order='F')

    # rmax =  rmax - 1
    # cmax = cmax - 1


    while (r <= rmax) and (c <= cmax):


        if np.mod(c + r, 2) == 0:    
            if r == rmin:
                output[i-1] = SI[r-1, c-1]
                if c == cmax:
                    r   = r + 1
                    sq_up_end = i
                    sq_down_begin = i+1
                    seq.append(output[sq_up_begin-1:sq_up_end])
                    j = j + 1

                else:
                    c = c + 1
                    sq_up_end = i
                    sq_down_begin = i+1

                    seq.append(output[sq_up_begin-1:sq_up_end])

                    j = j + 1

                i = i + 1

            elif (c == cmax) and (r < rmax):
                output[i-1] = SI[r-1, c-1]
                r = r + 1
                
                sq_up_end = i
                seq.append(output[sq_up_begin-1:sq_up_end])
                sq_down_begin =i+1
                j=j+1
                            
                i = i + 1

            elif (r > rmin) and (c < cmax):
                output[i-1] = SI[r-1, c-1]
                r = r - 1
                c = c + 1
                i = i + 1

        else:
            if (r == rmax) and (c <= cmax):
                output[i-1] = SI[r-1, c-1]
                c = c + 1
                sq_down_end = i
                seq.append(output[sq_down_begin-1:sq_down_end])
                sq_up_begin =i+1
                j = j + 1
                i = i + 1
            elif c == cmin:
                output[i-1] = SI[r-1, c-1]
                
                if r == rmax:
                    c = c + 1
                    
                    sq_down_end = i
                    seq.append(output[sq_down_begin-1:sq_down_end])
                    sq_up_begin =i+1
                    j = j + 1

                else:
                    r = r + 1
                    sq_down_end = i
                    seq.append(output[sq_down_begin-1:sq_down_end])
                    sq_up_begin =i+1
                    j = j + 1

                

                i = i + 1
            elif (r < rmax) and (c > cmin):
                output[i-1] = SI[r-1, c-1]
                r = r + 1
                c = c - 1
                i = i + 1


        if (r == rmax) and (c == cmax):
            output[i-1] = SI[r-1, c-1]
            sq_end = i
            seq.append(np.array([output[sq_end-1]]))

            break
    

    return seq


def fliplr(x):

    x = np.flip(x,1)

    return x

def getGLRLM(ROIonly,levels2):

    levels = levels2.copy()
    nLevel = len(levels)
    if nLevel > 100:
        adjust = 10000
    else:
        adjust = 1000
    


    levelTemp = np.max(levels)+1
    # ROIonly[np.isnan(ROIonly)] = levelTemp
    ROIonly = np.nan_to_num(ROIonly,nan = levelTemp)
    levels = np.append(levels, levelTemp)



    uniqueVol = np.round(levels*adjust)/adjust
    ROIonly=np.round(ROIonly*adjust)/adjust
    NL = len(levels) - 1



    sizeV = ROIonly.shape
    numInit = int(np.ceil(np.max(sizeV)*np.sqrt(3)))
    GLRLM = np.zeros((NL+1,numInit), dtype=np.float32)
    GLRLM_D = np.zeros((NL+1,numInit,13), dtype=np.float32)



    if ROIonly.ndim == 3:
        nComp = sizeV[2]
    else:
        nComp = 1
        ROIonly = np.expand_dims(ROIonly,axis=-1)
    
    
    for k in range (0,nComp):
        image = ROIonly[:,:,k].copy()
        uniqueIm = np.unique(image)
        NLtemp = len(uniqueIm)
        indexRow = np.zeros((NLtemp,1), dtype=np.float32)
        temp = image.copy()
        for j in range (0,NLtemp):
            indexRow[j] = np.where(uniqueIm[j]==uniqueVol)[0] + 1
            image[temp==uniqueIm[j]] = j+1
        
        GLRLMtemp = rle_0(image,NLtemp)
        # app = GLRLM()
        # glrlm = app.get_features(image,NLtemp)

        nRun = GLRLMtemp.shape[1]
        # nRun = GLRLMtemp.shape[1]

        ind = list(indexRow[0:NLtemp].flatten(order='F')  - 1)
        # ind = list(indexRow[0:NLtemp].flatten())
        ind = [int(item) for item in ind]

        GLRLM[ind,0:nRun] = GLRLM[ind,0:nRun] + GLRLMtemp[0:NLtemp,0:nRun]
        GLRLM_D[ind,0:nRun,0] = GLRLM_D[ind,0:nRun,0] + GLRLMtemp[0:NLtemp,0:nRun] 
        
        GLRLMtemp = rle_0( np.transpose(image),NLtemp)
        nRun = GLRLMtemp.shape[1]

        GLRLM[ind,0:nRun] = GLRLM[ind,0:nRun] + GLRLMtemp[0:NLtemp,0:nRun]
        GLRLM_D[ind,0:nRun,1] = GLRLM_D[ind,0:nRun,1] + GLRLMtemp[0:NLtemp,0:nRun] 
        

        seq = zigzag(image)
        GLRLMtemp = rle_45(seq,NLtemp)
        nRun = GLRLMtemp.shape[1]

        GLRLM[ind,0:nRun] = GLRLM[ind,0:nRun] + GLRLMtemp[0:NLtemp,0:nRun]
        GLRLM_D[ind,0:nRun,2] = GLRLM_D[ind,0:nRun,2] + GLRLMtemp[0:NLtemp,0:nRun] 
        

        seq = zigzag(fliplr(image))
        GLRLMtemp = rle_45(seq,NLtemp)
        nRun = GLRLMtemp.shape[1]

        GLRLM[ind,0:nRun] = GLRLM[ind,0:nRun] + GLRLMtemp[0:NLtemp,0:nRun]
        GLRLM_D[ind,0:nRun,3] = GLRLM_D[ind,0:nRun,3] + GLRLMtemp[0:NLtemp,0:nRun] 
        

    if len(np.squeeze(ROIonly).shape) == 3:

        nComp = sizeV[0]
        image = np.zeros((sizeV[2],sizeV[1]), dtype=np.float32)
        # print(np.mean(ROIonly))
        # print(np.nanmean(ROIonly))

        for k in range (0,nComp):
            for j in range (0,sizeV[2]):
                image[j,:] = ROIonly[k,:,j].copy()
            
            uniqueIm = np.unique(image)
            NLtemp = len(uniqueIm)
            indexRow = np.zeros((NLtemp,1), dtype=np.float32)
            temp = image.copy()
            for j in range(0,NLtemp):
                indexRow[j] = np.where(uniqueIm[j]==uniqueVol)[0] + 1
                image[temp==uniqueIm[j]] = j+1
            
            
            GLRLMtemp = rle_0( np.transpose(image),NLtemp)
            nRun = GLRLMtemp.shape[1]

            ind = list(indexRow[0:NLtemp].flatten(order='F')  - 1)
            ind = [int(item) for item in ind]
            
            GLRLM[ind,0:nRun] = GLRLM[ind,0:nRun] + GLRLMtemp[0:NLtemp,0:nRun]
            GLRLM_D[ind,0:nRun,4] = GLRLM_D[ind,0:nRun,4] + GLRLMtemp[0:NLtemp,0:nRun] 
        
            
            seq = zigzag(image)
            GLRLMtemp = rle_45(seq,NLtemp)
            nRun = GLRLMtemp.shape[1]

            GLRLM[ind,0:nRun] = GLRLM[ind,0:nRun] + GLRLMtemp[0:NLtemp,0:nRun]
            GLRLM_D[ind,0:nRun,5] = GLRLM_D[ind,0:nRun,5] + GLRLMtemp[0:NLtemp,0:nRun] 
        
            seq = zigzag(fliplr(image))
            GLRLMtemp = rle_45(seq,NLtemp)
            nRun = GLRLMtemp.shape[1]

            GLRLM[ind,0:nRun] = GLRLM[ind,0:nRun] + GLRLMtemp[0:NLtemp,0:nRun]
            GLRLM_D[ind,0:nRun,6] = GLRLM_D[ind,0:nRun,6] + GLRLMtemp[0:NLtemp,0:nRun] 


        nComp = sizeV[1]
        image = np.zeros((sizeV[0],sizeV[2]), dtype=np.float32)
        for k in range (0,nComp):
            for j in range (0,sizeV[2]):
                image[:,j] = ROIonly[:,k,j].copy()
            
            uniqueIm = np.unique(image)
            NLtemp = len(uniqueIm)
            indexRow = np.zeros((NLtemp,1), dtype=np.float32)
            temp = image.copy()
            for j in range (0,NLtemp):
                indexRow[j] = np.where(uniqueIm[j]==uniqueVol)[0] + 1
                image[temp==uniqueIm[j]] = j + 1
                        

            seq = zigzag(image)
            GLRLMtemp = rle_45(seq,NLtemp)
            nRun = GLRLMtemp.shape[1]

            ind = list(indexRow[0:NLtemp].flatten(order='F')  - 1)
            ind = [int(item) for item in ind]

            GLRLM[ind,0:nRun] = GLRLM[ind,0:nRun] + GLRLMtemp[0:NLtemp,0:nRun]
            GLRLM_D[ind,0:nRun,7] = GLRLM_D[ind,0:nRun,7] + GLRLMtemp[0:NLtemp,0:nRun] 
        

            seq = zigzag(fliplr(image))
            GLRLMtemp = rle_45(seq,NLtemp)
            nRun = GLRLMtemp.shape[1]
            GLRLM[ind,0:nRun] = GLRLM[ind,0:nRun] + GLRLMtemp[0:NLtemp,0:nRun]
            GLRLM_D[ind,0:nRun,8] = GLRLM_D[ind,0:nRun,8] + GLRLMtemp[0:NLtemp,0:nRun] 
        

        image = np.zeros((sizeV[2],sizeV[1]), dtype=np.float32)
        temp = np.random.rand(sizeV[2],sizeV[1])
        # d = np.arange(- (temp.shape[0]-1),temp.shape[1],1)
        # d = np.arange(- (temp.shape[0]-1),1,1)
        diagTemp = spdiags(temp)
        szDiag = diagTemp.shape
        diagMat1 = np.zeros((szDiag[0],szDiag[1],sizeV[0]), dtype=np.float32)
        diagMat2 = np.zeros((diagTemp.shape[0],diagTemp.shape[1],sizeV[0]), dtype=np.float32)
        for k in range (0,sizeV[0]):
            for j in range (0,sizeV[2]):
                current_ROI = ROIonly[k,:,j].copy()
                image[j,:] = current_ROI
            
            try:
                # d = np.arange(- (image.shape[0]-1),image.shape[1],1)
                diagMat1[:,:,k]=spdiags(image)
                # print(diagMat1[:,:,k])
            except:
                # d = np.arange(- (image.shape[0]-1),image.shape[1],1)
                temp=spdiags(image)
                # print(temp)
                numberDiff=np.abs(temp.shape[1] - diagMat1.shape[1])
                if np.mod(numberDiff,2):
                    temp=np.pad(temp,pad_width=(  (0,0),((numberDiff+1)/2,(numberDiff+1)/2),(0,0)   ),mode='constant',constant_values = 0)
                    diagMat1[:,:,k]=temp[:,:-1].copy()
                else:
                    diagMat1[:,:,k]=np.pad(temp,pad_width=(  (0,0),(numberDiff/2,numberDiff/2),(0,0)   ),mode='constant',constant_values = 0)
                
            
            try:
                # d = np.arange(- (image.shape[0]-1),image.shape[1],1)

                diagMat2[:,:,k]=spdiags(fliplr(image))
                # print(diagMat2[:,:,k])
            except:
                # d = np.arange(- (image.shape[0]-1),image.shape[1],1)
                temp = spdiags(fliplr(image))
                # print(temp)
                numberDiff = np.abs(temp.shape[1]-diagMat2.shape[1])
                if np.mod(numberDiff,2):
                    temp = np.pad(temp,pad_width=(  (0,0),((numberDiff+1)/2,(numberDiff+1)/2),(0,0)   ),mode='constant',constant_values = 0)
                    diagMat2[:,:,k] = temp[:,:-1].copy()
                else:
                    diagMat2[:,:,k] = np.pad(temp,pad_width=(  (0,0),(numberDiff/2,numberDiff/2),(0,0)   ),mode='constant',constant_values = 0)

        for j in range (0,szDiag[1]):
            index = np.where(diagMat1[:,j,0]!=0)[0]
            nTemp = len(index)
            image1 = np.zeros((sizeV[0],nTemp), dtype=np.float32)
            image2 = np.zeros((sizeV[0],nTemp), dtype=np.float32)
            for k in range (0,sizeV[0]):
                image1[k,:nTemp] = np.transpose(diagMat1[index,j,k])
                image2[k,:nTemp] = np.transpose(diagMat2[index,j,k])
            
            
            uniqueIm = np.unique(image1)
            NLtemp = len(uniqueIm)
            indexRow = np.zeros((NLtemp,1), dtype=np.float32)
            temp = image1.copy()
            for k in range(0,NLtemp):
                indexRow[k] = np.where(uniqueIm[k]==uniqueVol)[0] + 1
                image1[temp==uniqueIm[k]] = k + 1
            
                        

            seq = zigzag(image1)
            GLRLMtemp = rle_45(seq,NLtemp)

            nRun = GLRLMtemp.shape[1]

            ind = list(indexRow[0:NLtemp].flatten(order='F')  - 1)
            ind = [int(item) for item in ind]

            GLRLM[ind,0:nRun] = GLRLM[ind,0:nRun] + GLRLMtemp[0:NLtemp,0:nRun]
            GLRLM_D[ind,0:nRun,9] = GLRLM_D[ind,0:nRun,9] + GLRLMtemp[0:NLtemp,0:nRun] 

            seq = zigzag(fliplr(image1))
            GLRLMtemp = rle_45(seq,NLtemp)
            nRun = GLRLMtemp.shape[1]
            GLRLM[ind,0:nRun] = GLRLM[ind,0:nRun] + GLRLMtemp[0:NLtemp,0:nRun]
            GLRLM_D[ind,0:nRun,10] = GLRLM_D[ind,0:nRun,10] + GLRLMtemp[0:NLtemp,0:nRun] 


            uniqueIm = np.unique(image2)
            NLtemp = len(uniqueIm)
            indexRow = np.zeros((NLtemp,1), dtype=np.float32)
            temp = image2.copy()
            for k in range(0,NLtemp):
                indexRow[k] = np.where(uniqueIm[k]==uniqueVol)[0] + 1
                image2[temp==uniqueIm[k]] = k +1 
            
            seq = zigzag(image2)
            GLRLMtemp = rle_45(seq,NLtemp)
            nRun = GLRLMtemp.shape[1]

            ind = list(indexRow[0:NLtemp].flatten(order='F')  - 1)
            ind = [int(item) for item in ind]

            GLRLM[ind,0:nRun] = GLRLM[ind,0:nRun] + GLRLMtemp[0:NLtemp,0:nRun]
            GLRLM_D[ind,0:nRun,11] = GLRLM_D[ind,0:nRun,11] + GLRLMtemp[0:NLtemp,0:nRun] 

            seq = zigzag(fliplr(image2))
            GLRLMtemp = rle_45(seq,NLtemp)
            nRun = GLRLMtemp.shape[1]
            GLRLM[ind,0:nRun] = GLRLM[ind,0:nRun] + GLRLMtemp[0:NLtemp,0:nRun]
            GLRLM_D[ind,0:nRun,12] = GLRLM_D[ind,0:nRun,12] + GLRLMtemp[0:NLtemp,0:nRun] 


    GLRLM = GLRLM[:-1,:]
    sumGLRLM = np.sum(GLRLM,axis=0)
    
    non_zero_indices = np.where(sumGLRLM > 0)[0]
    if len(non_zero_indices) == 0:
        # No valid run-length patterns found - return minimal matrix
        import logging
        logging.getLogger("Dev_logger").warning(f"No valid run-length patterns found in GLRLM. ROI may be too small or have insufficient texture complexity.")
        GLRLM = np.zeros((GLRLM.shape[0], 1), dtype=np.float32)  # Minimal 1-column matrix
        stop = 0
    else:
        stop = int(non_zero_indices[-1])
    
    GLRLM = GLRLM[:,:stop+1]

    GLRLM_D = GLRLM_D[:-1,:,:]
    GLRLM_D = GLRLM_D[:,:GLRLM.shape[1],:]


    return GLRLM , GLRLM_D



def ShortRunEmph(GLRLM,nG,nR,nS):
    Rj = np.sum(GLRLM , axis=0)
    J = np.arange(0,nR) 
    J = J +1
    tmp = np.divide (Rj , np.float_power(J,2))
    f_SRE = np.sum(tmp)/nS
    return f_SRE

def LongRunEmph(GLRLM,nG,nR,nS):
    Rj =np.sum(GLRLM , axis=0)
    J = np.arange(0,nR)
    J = J +1
    tmp = np.multiply (Rj , np.float_power(J,2))
    f_LRE = np.sum(tmp)/nS
    return f_LRE

def LowGLRunRmph(GLRLM,nG,nR,nS):

    Ri = np.sum(np.atleast_2d(GLRLM) , axis=1)
    I = np.transpose (np.arange(0,nG))
    I = I +1
    tmp = np.divide(Ri , np.float_power(I,2))
    f_LGLRE = np.sum(tmp)/nS
    return f_LGLRE

def HighGLRunEmph(GLRLM,nG,nR,nS):
    Ri = np.sum(np.atleast_2d(GLRLM) , axis=1)
    I = np.transpose (np.arange(0,nG))
    I = I +1
    tmp = np.multiply(Ri , np.float_power(I,2))
    f_HGLRE = np.sum(tmp)/nS
    return f_HGLRE

def ShortRunLowGLEmph(GLRLM,nG,nR,nS):

    rangeX = np.arange(1, nG+1 , 1)
    rangeY = np.arange(1, nR+1 , 1)

    I , J = np.meshgrid(rangeX,rangeY)
    tmp = np.divide (np.transpose(GLRLM) , (  np.multiply(  np.float_power (I,2) , np.float_power (J,2))      ))
    f_SRLGLE = np.sum(tmp)/nS

    return f_SRLGLE

def ShortRunHighRL(GLRLM,nG,nR,nS):

    rangeX = np.arange(1, nG+1 , 1)
    rangeY = np.arange(1, nR+1 , 1)
    I , J = np.meshgrid(rangeX,rangeY)
    tmp = np.divide(np.multiply (np.float_power(I,2), np.transpose(GLRLM)) , np.float_power(J,2))
    f_SRHGLE = np.sum(tmp)/nS

    return f_SRHGLE

def LongRunLowGLEmph(GLRLM,nG,nR,nS):

    rangeX = np.arange(1, nG+1 , 1)
    rangeY = np.arange(1, nR+1 , 1)
    I , J = np.meshgrid(rangeX,rangeY)
        
    tmp = np.divide(np.multiply (np.float_power(J,2),np.transpose(GLRLM)) , np.float_power(I,2))
    f_LRLGLE = np.sum(tmp)/nS
    return f_LRLGLE

def LongRunHighGLEmph(GLRLM,nG,nR,nS):

    rangeX = np.arange(1, nG+1 , 1)
    rangeY = np.arange(1, nR+1 , 1)
    I , J = np.meshgrid(rangeX,rangeY)
    
    tmp = np.multiply(np.multiply (np.float_power(J,2),np.float_power(I,2)) ,np.transpose(GLRLM) )

    f_LRHGLE = np.sum(tmp)/nS

    return f_LRHGLE

def GLnonUnif(GLRLM,nS):


    Ri = np.sum(np.atleast_2d(GLRLM) , axis=1)
    f_GLNU = np.sum( np.float_power(Ri,2) ) / nS

    return f_GLNU


def GLnonUnifNormzd(GLRLM,nS):

    Ri = np.sum(np.atleast_2d(GLRLM) , axis=1)
    f_GLNUN = np.sum( np.float_power(Ri,2)) / np.float_power(nS,2)
    return f_GLNUN


def RunLengthNonUnif(GLRLM,nS):


    Ri = np.sum(GLRLM , axis=0)
    f_RLNU = np.sum(np.float_power(Ri,2)) / nS

    return f_RLNU


def RunLengthNonUnifNormzd(GLRLM,nS):

    Ri = np.sum(GLRLM , axis=0)
    f_RLNUN = np.sum( np.float_power(Ri,2)) / np.float_power(nS,2)
    return f_RLNUN


def RunPercentage(nS,nV):


    f_RP = nS / nV
    return f_RP


def GLVar(GLRLM,nG,nR,nS):

    Pij = GLRLM / nS

    rangeX = np.arange(1, nG+1 , 1)
    rangeY = np.arange(1, nR+1 , 1)
    I , J = np.meshgrid(rangeX,rangeY)

    mu = np.sum(np.sum(np.multiply(I, np.transpose(Pij))))
    tmp =  np.multiply (np.float_power((I - mu),2), np.transpose(Pij))
    f_GLV = np.sum(tmp)

    return f_GLV


def RunLengthVar(GLRLM,nG,nR,nS):

    Pij = GLRLM / nS
    rangeX = np.arange(1, nG+1 , 1)
    rangeY = np.arange(1, nR+1 , 1)
    I , J = np.meshgrid(rangeX,rangeY)
    
    mu = np.sum(np.sum(np.multiply(J, np.transpose(Pij))))
    tmp =  np.multiply (np.float_power((J - mu),2), np.transpose(Pij))
    f_RLV = np.sum(tmp)
    return f_RLV


def RunEntropy(GLRLM,nS, feature_value_mode='REAL_VALUE', epsilon=1e-30):

    Pij = GLRLM / nS
    tmp = np.multiply(Pij , np.log2(Pij+ epsilon))
    f_RE = -np.sum(tmp)

    return f_RE






















def GLRLMfeatHandler(GLRLM,nG,nR,nS,nV, feature_value_mode='REAL_VALUE'):

    
    try:
        nf = GLRLM.shape[1]
    except:
        GLRLM = np.expand_dims(GLRLM,-1)

    if nS == 0:
        value = np.nan
        import logging
        output = [value] * 16
        logging.getLogger("Dev_logger").warning(f"Empty GLRLM matrix (nS=0). ROI may be too small or empty or it has only 1 voxel. Returning {value} values.")
        return output

    ArrayOut = []
    ArrayOut.append(ShortRunEmph(GLRLM,nG,nR,nS))
    ArrayOut.append(LongRunEmph(GLRLM,nG,nR,nS))
    ArrayOut.append(LowGLRunRmph(GLRLM,nG,nR,nS))
    ArrayOut.append(HighGLRunEmph(GLRLM,nG,nR,nS))
    ArrayOut.append(ShortRunLowGLEmph(GLRLM,nG,nR,nS))
    ArrayOut.append(ShortRunHighRL(GLRLM,nG,nR,nS))
    ArrayOut.append(LongRunLowGLEmph(GLRLM,nG,nR,nS))
    ArrayOut.append(LongRunHighGLEmph(GLRLM,nG,nR,nS))
    ArrayOut.append(GLnonUnif(GLRLM,nS))
    ArrayOut.append(GLnonUnifNormzd(GLRLM,nS))
    ArrayOut.append(RunLengthNonUnif(GLRLM,nS))
    ArrayOut.append(RunLengthNonUnifNormzd(GLRLM,nS))
    ArrayOut.append(RunPercentage(nS,nV))
    ArrayOut.append(GLVar(GLRLM,nG,nR,nS))
    ArrayOut.append(RunLengthVar(GLRLM,nG,nR,nS))
    ArrayOut.append(RunEntropy(GLRLM,nS, feature_value_mode=feature_value_mode))

    return ArrayOut


def CalcGLRLM(GLRLM,ROI, feature_value_mode='REAL_VALUE'):

    
    nG = GLRLM.shape[0]            
    nR = GLRLM.shape[1]               

    nV = np.where(~np.isnan(ROI))[0].shape[0]
    # nV = np.where(~np.isnan(ROI))[0]

    FeatMatrixout = []

    if GLRLM.ndim != 3:
        maxRange = 1
        tmpGLRLM = np.squeeze(GLRLM)
        nS = np.sum(tmpGLRLM)        
        
        feat = np.transpose(GLRLMfeatHandler(tmpGLRLM,nG,nR,nS,nV, feature_value_mode=feature_value_mode))
        FeatMatrixout = feat

    elif GLRLM.ndim == 3:
        maxRange = GLRLM.shape[2]
        for k in range(0,maxRange):
            tmpGLRLM = np.squeeze(GLRLM[:,:,k])
            nS = np.sum(tmpGLRLM)   
            feat = np.transpose(GLRLMfeatHandler(tmpGLRLM,nG,nR,nS,nV, feature_value_mode=feature_value_mode))

            if k == 0:
                FeatMatrixout = feat
            else:    
                FeatMatrixout = np.column_stack((FeatMatrixout , feat))
            
    return FeatMatrixout



import numpy as np
import scipy
import cmath
import gc
# from skimage.measure import marching_cubes
from ..visera.SERAutilities import *

from pysera.utils.utils import handle_math_operations
from ...processing.synthesize_RoIs import synthesis_small_RoI

import logging

logger = logging.getLogger("Dev_logger")

# -------------------------------------------------------------------------
# function [MorphVect] = getMorph(MorphBox,MorphROI,ROIints, pixelW,sliceS)
# -------------------------------------------------------------------------
# DESCRIPTION:
# This function calculates morphological features.
# -------------------------------------------------------------------------
# INPUTS:
# - MorphBox: The smallest box containing the resampled 3D ROI, with the
#             imaging data ready for texture analysis computations. Voxels
#             outside the ROI are set to NaNs.
# - MorphROI: The smallest box containing the 3D morphological ROI. Voxels
#             outside the ROI are set to 0. It has only 0 and 1.
# - ROIints: The smallest box containing the 3D intensity ROI with
#             their intensities. Voxels outside the ROI are set to NaNs.
# - pixelW: width of the voxel in the X (=Y) direction
# - sliceS: Slice thickness of the voxel in the Z direction
#
# Note: The first 3 parameters are outputs of prepareVolume function.
# -------------------------------------------------------------------------
# OUTPUTS:
# - MorphVect: An array of calculated morphological features
# -------------------------------------------------------------------------
# AUTHOR(S):
# - Saeed Ashrafinia
# - Mahdi Hosseinzadeh
# -------------------------------------------------------------------------
# HISTORY:
# - Creation: December 2022
# -------------------------------------------------------------------------


def getMorph(MorphBox2, _MorphROI2, ROIints2, pixelW, sliceS, feature_value_mode='REAL_VALUE'):
    MorphVect = []
    
    # Load roi from disk
    MorphROI = _MorphROI2.copy()
    ROIints = ROIints2.copy()
    # MorphBox = MorphBox2.copy()
    # MorphROI[MorphROI == 0] = np.nan
    # MorphROI_disk = np.load(MorphROI2_path, mmap_mode='c')
    morph_count = np.count_nonzero(MorphROI)
    RoI_count = np.count_nonzero(ROIints2)

    if morph_count < 2 and feature_value_mode=='APPROXIMATE_VALUE':             # Synthesize 1 voxels
            synth_MorphROI = synthesis_small_RoI(morph_count, MorphROI.copy(), background=0, target_num=2)
            synth_ROIints = synthesis_small_RoI(RoI_count, ROIints.copy(), background='NaN', target_num=2)
            synth_flag = True
    else:
        # if feature_value_mode=='REAL_VALUE' and NOT morph_count < 2
        synth_MorphROI = None
        synth_ROIints = None
        synth_flag = False

    # Clean RAM
    del ROIints2
    gc.collect()

    ROIboxF, ROIboxV, _ = getMesh(MorphROI, pixelW, pixelW, sliceS)
    MV, Surface = getVolSurfaceMesh(ROIboxF, ROIboxV)

    # from skimage.measure import mesh_surface_area
    # ROIboxV = ROIboxV - 1
    # ROIboxF = ROIboxF - 1
    # area = mesh_surface_area(verts=ROIboxV.astype(np.uint32), faces=ROIboxF.astype(np.uint32))

    MorphROI[MorphROI == 0] = np.nan
    if synth_flag:
        MVappx = morph_count
    else:
        MVappx = getVolume(MorphROI, pixelW, sliceS)

    xM, yM, zM = ind2sub(MorphROI)
    xI, yI, zI = ind2sub(ROIints)
    try:
        concat = np.column_stack((xM * pixelW, yM * pixelW, zM * sliceS)) + 1
        concat2 = np.column_stack((xI * pixelW, yI * pixelW, zI * sliceS)) + 1
        aa = np.nanmean(concat, axis=0)
        ROIints_flatten = ROIints.flatten(order='F')
        notnanval = np.where(~np.isnan(ROIints_flatten))
        arr = np.expand_dims(ROIints_flatten[notnanval[0]], -1)
        tile = np.tile(arr, reps=(1, 3))
        bb = np.nansum(np.multiply(concat2, tile), axis=0)
        cc = np.tile(np.nansum(ROIints), (1, 3))
        dd = aa - np.divide(bb, cc)

        CoM_shift = np.linalg.norm(dd)

    except:
        print('**** ROI is 1D. Brace for problems.....')
        CoM_shift = np.abs(np.mean(np.squeeze(MorphROI) - np.squeeze(ROIints)))

    Shape = getShape(MV, Surface)
    compactness1 = Shape['compactness1']
    compactness2 = Shape['compactness2']
    sphericity = Shape['sphericity']
    SVratio = Shape['SVratio']
    Irregularity = Shape['Irregularity']
    sphericalDisp = Shape['sphericalDisp']
    Asphericity = Shape['Asphericity']

    # Volume Densities
    VDs, ConvHullV = getVDs(MorphROI, synth_MorphROI, synth_flag, ROIboxV, pixelW, sliceS, MV, Surface, feature_value_mode=feature_value_mode)
    Mean = np.nanmean(ROIints)
    TLG = Mean * MV

    ## 1111111
    ROIints = ROIints.flatten(order='F')
    MorphROI3 = MorphROI.flatten(order='F')

    MorphROI3 = MorphROI3[~ np.isnan(ROIints)]
    ROIints = ROIints[~ np.isnan(ROIints)]
    ROIints = ROIints[~ np.isnan(MorphROI3)]

    xI = xI[~ np.isnan(MorphROI3)]
    yI = yI[~ np.isnan(MorphROI3)]
    zI = zI[~ np.isnan(MorphROI3)]


    ## 22222
    # ROIints_flatten = ROIints.flatten(order='F')
    # MorphROI3 = MorphROI.flatten(order='F')

    # print(np.mean(ROIints))
    # ROIints = ROIints[~np.isnan(MorphROI)]
    # MorphROI3 = MorphROI2.flatten(order='F')
    # xI = xI[MorphROI3 != 0]
    # yI = yI[MorphROI3 != 0]
    # zI = zI[MorphROI3 != 0]

    # # MoranI, GearyC = getAutoCorrs(ROIints, xI,yI,zI , pixelW,sliceS , Mean)
    # if MoranI == None and Max3Ddiam == None and feature_value_mode=='REAL_VALUE':
    #     Max3Ddiam = np.nan
    #     MoranI, GearyC = np.nan, np.nan
    # else:
    Max3Ddiam = getMax3Ddiam(ROIboxV, ConvHullV)
    MoranI, GearyC = getAutoCorrs(ROIints, xI, yI, zI, pixelW, sliceS)
    # MoranI, GearyC = getAutoCorrs(ROIints_flatten, xI,yI,zI,pixelW,sliceS )

    threshold = 2
    # GT2 = getVolume_gt(MorphBox,threshold,pixelW,sliceS)
    MorphVect = [MV, MVappx, Surface, SVratio, compactness1, compactness2,
                 sphericalDisp, sphericity, Asphericity, CoM_shift, Max3Ddiam]
    for v in VDs:
        MorphVect.append(v)

    MorphVect.append(TLG)
    MorphVect.append(MoranI)
    MorphVect.append(GearyC)

    return MorphVect


def getMesh(ROIbox, pixelWx, pixelWy, sliceS):
    dx, dy, dz = ROIbox.shape

    morphology_mask = np.pad(ROIbox, pad_width=((1, 1), (1, 1), (1, 1)), mode="constant", constant_values=0.0)

    # Ensure coordinate arrays match the exact dimensions of the morphology_mask
    mask_dx, mask_dy, mask_dz = morphology_mask.shape

    rangeX = np.linspace(0, pixelWx * (mask_dx - 1), mask_dx)
    rangeY = np.linspace(0, pixelWy * (mask_dy - 1), mask_dy)
    rangeS = np.linspace(0, sliceS * (mask_dz - 1), mask_dz)

    X, Y, Z = np.meshgrid(rangeX, rangeY, rangeS, indexing='ij')

    faces, vertices, norms = MarchingCubes(X, Y, Z, morphology_mask, 0.5, None)

    # from skimage.measure import marching_cubes

    # # Get ROI and pad with empty voxels
    # morphology_mask = np.pad(ROIbox, pad_width=1, mode="constant", constant_values=0.0)

    # # Use marching cubes to generate a mesh grid for the ROI
    # vertices, faces, norms, values = marching_cubes(volume=morphology_mask, level=0.5, spacing=(float(pixelWx),float(pixelWy),float(sliceS)))

    return faces, vertices, norms


def getVolSurfaceMesh(F, V):
    nFaces = F.shape[0]
    vols = np.zeros((nFaces, 1), dtype=np.float32)
    meanMat = np.mean(V, axis=0)
    vertices = V - meanMat
    F = F.astype(np.int64)
    for i in range(0, nFaces):
        ind = list(F[i, :] - 1)
        tetra = vertices[ind, :]

        vols[i] = np.linalg.det(tetra) / 6

    vol_out = np.abs(np.sum(vols))

    a = V[list(F[:, 1] - 1), :] - V[list(F[:, 0] - 1), :]
    b = V[list(F[:, 2] - 1), :] - V[list(F[:, 0] - 1), :]
    c = np.cross(a, b)
    area_out = 1 / 2 * np.sum(np.sqrt(np.sum(np.float_power(c, 2), axis=1)))

    return vol_out, area_out


def getVolume(ROIonly, pixelW, sliceS):
    # mask = ROIonly[~isnan(ROIonly)]
    numberVoxel = np.nansum(ROIonly.astype(np.float64))
    volume = numberVoxel * pixelW * pixelW * sliceS

    if hasattr(volume, '__len__') == True:
        volume = np.squeeze(volume)

    return volume


def getShape(V, S):
    compactness1 = V / np.sqrt(np.pi) / np.sqrt(np.float_power(S, 3))
    compactness2 = 36 * np.pi * np.float_power(V, 2) / np.float_power(S, 3)
    sphericity = np.float_power((36 * np.pi * np.float_power(V, 2)), (1 / 3)) / S
    sphericalDisp = S / np.float_power((36 * np.pi * np.float_power(V, 2)), (1 / 3))
    Asphericity = np.float_power((np.float_power(S, 3) / np.float_power(V, 2) / 36 / np.pi), (1 / 3)) - 1
    SVratio = S / V
    Scircle = np.float_power((3 * V / (4 * np.pi)), (2 / 3)) * 4 * np.pi
    Irregularity = S / Scircle

    Shape = {}
    Shape['compactness1'] = compactness1
    Shape['compactness2'] = compactness2
    Shape['sphericity'] = sphericity
    Shape['SVratio'] = SVratio
    Shape['Irregularity'] = Irregularity
    Shape['sphericalDisp'] = sphericalDisp
    Shape['Asphericity'] = Asphericity

    return Shape


def getVDs(ROIbox, synth_ROIbox, synth_flag, ROIboxV, pixelW, sliceS, MV, Surface, epsilon=1e-30, feature_value_mode='REAL_VALUE'):
    if synth_flag:
        x, y, z = ind2sub(synth_ROIbox)
    else:
        x, y, z = ind2sub(ROIbox)

    concat = np.column_stack((x * pixelW, y * pixelW, z * sliceS)) + 1

    covariance = np.cov(np.transpose(concat))
    MaxV = 20
    v0 = np.arange(0, MaxV + 1, 1)

    # Get eighenvalues
    # print(covariance, np.all(np.isnan(covariance)), 54678437685478365)
    if np.all(np.isnan(covariance)):
        MALs = np.full((3,), np.nan)
        Elong, Fltns = np.nan, np.nan
        VDaee, ADaee = np.nan, np.nan
        if feature_value_mode=='APPROXIMATE_VALUE':
            logger.error(f"Issue with synthesis. Returning NaN for MALs, Elong, Fltns, VDaee, ADaee.")
    else:
        EIGs = np.linalg.eigh(covariance)[0]  # EIGs =  scipy.linalg.eigh(covariance)[0]

        EIGs = np.transpose(np.flip(EIGs))

        EIGs = handle_math_operations(EIGs, feature_value_mode, 'sqrt', epsilon)

        if EIGs.shape[0] == 1:
            EIGs = np.tile(EIGs, (1, 3))

        L1 = 2 * np.sqrt(EIGs[0])
        L2 = 2 * np.sqrt(EIGs[1])
        L3 = 2 * np.sqrt(EIGs[2])

        MALs = 4 * np.sqrt(EIGs)
        # EIGs[0] = handle_math_operations(np.array(EIGs[0]), feature_value_mode, 'divide', epsilon)
        if feature_value_mode=='APPROXIMATE_VALUE' and EIGs[0] == 0.:
            Elong = np.sqrt(EIGs[1] / epsilon)
            Fltns = np.sqrt(EIGs[2] / epsilon)
        else:
            Elong = np.sqrt(EIGs[1] / EIGs[0])
            Fltns = np.sqrt(EIGs[2] / EIGs[0])

        if feature_value_mode=='APPROXIMATE_VALUE' and np.prod(np.sqrt(EIGs)) == 0.:
            VDaee = 3 * MV / (4 * np.pi * np.float_power(2, 3) * np.prod(np.sqrt(handle_math_operations(EIGs, feature_value_mode, 'divide', epsilon))))
        else:
            VDaee = 3 * MV / (4 * np.pi * np.float_power(2, 3) * np.prod(np.sqrt(EIGs)))

        alpha = np.sqrt(1 - np.float_power(L2, 2) / np.float_power(L1, 2))
        beta = np.sqrt(1 - np.float_power(L3, 2) / np.float_power(L1, 2))


        sumtmp = []
        aa = (np.float_power(alpha, 2) + np.float_power(beta, 2)) / (2 * alpha * beta)
        mm = np.min([1, aa])
        for k in range(0, MaxV + 1):

            bb = np.transpose(np.array(legendre(k, np.array(mm))))
            cc = (1 - 4 * np.float_power(k, 2))
            ee = np.float_power((alpha * beta), k)
            ff = np.divide(ee, cc)
            dd = np.multiply(ff, bb)
            if np.isscalar(dd):
                tem = dd
                dd = np.array([tem])
            else:
                dd = dd[0]
            sumtmp.append(dd)

        Final_sumtmp = []
        for i in range(0, len(sumtmp)):  # toto: use extend instead ???
            for j in sumtmp[i]:
                Final_sumtmp.append(j)

        sumtmp = np.array(Final_sumtmp)
        aee = 4 * np.pi * L1 * L2 * np.sum(sumtmp)
        aee = handle_math_operations(np.array(aee), feature_value_mode, 'divide', epsilon)
        ADaee = Surface / aee

    # if synth_flag:
    #     nx, ny, nz = synth_ROIbox.shape
    # else:
    nx, ny, nz = ROIbox.shape

    VDaabb = MV / np.prod(np.multiply(np.array([nx, ny, nz]), np.transpose(np.array([pixelW, pixelW, sliceS]))))
    ADaabb = (Surface / (2 * nx * pixelW * ny * pixelW + 2 * ny * pixelW * nz * sliceS + 2 * nx * pixelW * nz * sliceS))

    if hasattr(ADaabb, '__len__') == True:
        ADaabb = np.squeeze(ADaabb)

    covariance = np.cov(np.transpose(ROIboxV))
    EIGs = np.linalg.eigh(covariance)[0]
    # EIGs = np.sum(EIGs , axis = 0)
    L1 = 2 * np.sqrt(EIGs[0])
    L2 = 2 * np.sqrt(EIGs[1])
    L3 = 2 * np.sqrt(EIGs[2])
    VDmvee = MV / (4 * np.pi * L1 * L2 * L3 / 3)

    aa = (np.float_power(L2, 2) / np.float_power(L1, 2))
    bb = (np.float_power(L3, 2) / np.float_power(L1, 2))
    alpha = cmath.sqrt(1 - aa)
    beta = cmath.sqrt(1 - bb)

    try:
        yy = 4 * np.pi * L1 * L2

        aa = (np.float_power(alpha, 2) + np.float_power(beta, 2)) / (2 * alpha * beta)
        mm = np.min([1, aa])
        bb = np.transpose(np.array(legendre(MaxV, np.array(mm))))
        cc = (1 - 4 * np.float_power(v0, 2))
        ee = np.float_power((alpha * beta), v0)
        ff = np.divide(ee, cc)
        dd = np.multiply(ff, bb)

        if np.isscalar(dd):
            tem = dd
            dd = np.array([tem])
        else:
            dd = dd[0]

        aee = yy * np.sum(dd)

    except:
        EIGs = np.transpose(np.flip(EIGs))

        L1 = 2 * np.sqrt(EIGs[0])
        L2 = 2 * np.sqrt(EIGs[1])
        L3 = 2 * np.sqrt(EIGs[2])

        VDmvee = MV / (4 * np.pi * L1 * L2 * L3 / 3)

        alpha = cmath.sqrt(1 - (np.float_power(L2, 2) / np.float_power(L1, 2)))
        beta = cmath.sqrt(1 - (np.float_power(L3, 2) / np.float_power(L1, 2)))

        yy = 4 * np.pi * L1 * L2

        aa = (np.float_power(alpha, 2) + np.float_power(beta, 2)) / (2 * alpha * beta)
        mm = np.min([1, aa])
        bb = np.transpose(np.array(legendre(MaxV, np.array(mm))))
        cc = (1 - 4 * np.float_power(v0, 2))
        ee = np.float_power((alpha * beta), v0)
        ff = np.divide(ee, cc)
        dd = np.multiply(ff, bb)

        if np.isscalar(dd):
            tem = dd
            dd = np.array([tem])
        else:
            dd = dd[0]
        aee = yy * np.sum(dd)
    aee = np.real(aee)

    if aee == 0. and feature_value_mode == 'APPROXIMATE_VALUE':
        ADmvee = Surface / epsilon
    elif aee == 0. and feature_value_mode == 'REAL_VALUE':
        ADmvee = np.nan
    else:
        ADmvee = Surface / aee

    try:
        _, _, Vombb, Aombb, _ = minboundbox(x * pixelW, y * pixelW, z * sliceS, 'volume', 1, feature_value_mode=feature_value_mode)
    except:
        print('This is a 2D or 1D ROI. Switch to 2D Convex Hull and Bounding Box calculation.')
        if synth_flag:
            newROIbox = np.squeeze(synth_ROIbox)
        else:
            newROIbox = np.squeeze(ROIbox)

        xtmp, ytmp, _ = ind2sub(newROIbox)
        try:
            if synth_flag:
                RoIbox_shape = synth_ROIbox.shape
            else:
                RoIbox_shape = ROIbox.shape

            if RoIbox_shape[2] == 1:
                _, Aombb, Vombb = minBoundingBox2D(xtmp, ytmp, pixelW, pixelW, sliceS, feature_value_mode=feature_value_mode)
            elif RoIbox_shape[1] == 1:
                _, Aombb, Vombb = minBoundingBox2D(xtmp, ytmp, pixelW, sliceS, pixelW, feature_value_mode=feature_value_mode)
            elif RoIbox_shape[0] == 1:
                _, Aombb, Vombb = minBoundingBox2D(xtmp, ytmp, sliceS, pixelW, pixelW, feature_value_mode=feature_value_mode)
            else:
                print('Min bounding box does not respond (no Convex Hull available). Set OMBB = AABB....')
                arr1 = np.column_stack((nx, ny, nz))
                arr2 = np.column_stack((pixelW, pixelW, sliceS))

                Vombb = np.prod(np.multiply(arr1, arr2))
                Aombb = 2 * nx * pixelW * ny * pixelW + 2 * ny * pixelW * nz * sliceS + 2 * nx * pixelW * nz * sliceS


        except:
            print('Min bounding box does not respond (probably a 1D ROI). Set OMBB = AABB....')
            arr1 = np.column_stack((nx, ny, nz))
            arr2 = np.column_stack((pixelW, pixelW, sliceS))
            Vombb = np.prod(np.multiply(arr1, arr2))
            Aombb = 2 * nx * pixelW * ny * pixelW + 2 * ny * pixelW * nz * sliceS + 2 * nx * pixelW * nz * sliceS

    VDombb = MV / Vombb
    ADombb = Surface / Aombb

    # approache 1
    K2 = scipy.spatial.ConvexHull(ROIboxV)
    Vch = K2.volume
    K = K2.simplices

    # approache 2
    # hull = pyhull.convex_hull.ConvexHull(ROIboxV)
    # K = [x for x in hull.vertices if len(x) == 3]
    # K = np.array(K)

    Solidity = MV / Vch

    CHarea = np.sum(np.sqrt(np.sum(
        np.float_power(
            (
                    np.column_stack((
                        np.multiply(ROIboxV[K[:, 0], 1], ROIboxV[K[:, 1], 2]) - np.multiply(ROIboxV[K[:, 0], 2],
                                                                                            ROIboxV[K[:, 1], 1]),
                        np.multiply(ROIboxV[K[:, 0], 2], ROIboxV[K[:, 1], 0]) - np.multiply(ROIboxV[K[:, 0], 0],
                                                                                            ROIboxV[K[:, 1], 2]),
                        np.multiply(ROIboxV[K[:, 0], 0], ROIboxV[K[:, 1], 1]) - np.multiply(ROIboxV[K[:, 0], 1],
                                                                                            ROIboxV[K[:, 1], 0])))
                    +
                    np.column_stack((
                        np.multiply(ROIboxV[K[:, 1], 1], ROIboxV[K[:, 2], 2]) - np.multiply(ROIboxV[K[:, 1], 2],
                                                                                            ROIboxV[K[:, 2], 1]),
                        np.multiply(ROIboxV[K[:, 1], 2], ROIboxV[K[:, 2], 0]) - np.multiply(ROIboxV[K[:, 1], 0],
                                                                                            ROIboxV[K[:, 2], 2]),
                        np.multiply(ROIboxV[K[:, 1], 0], ROIboxV[K[:, 2], 1]) - np.multiply(ROIboxV[K[:, 1], 1],
                                                                                            ROIboxV[K[:, 2], 0])))
                    +
                    np.column_stack((
                        np.multiply(ROIboxV[K[:, 2], 1], ROIboxV[K[:, 0], 2]) - np.multiply(ROIboxV[K[:, 2], 2],
                                                                                            ROIboxV[K[:, 0], 1]),
                        np.multiply(ROIboxV[K[:, 2], 2], ROIboxV[K[:, 0], 0]) - np.multiply(ROIboxV[K[:, 2], 0],
                                                                                            ROIboxV[K[:, 0], 2]),
                        np.multiply(ROIboxV[K[:, 2], 0], ROIboxV[K[:, 0], 1]) - np.multiply(ROIboxV[K[:, 2], 1],
                                                                                            ROIboxV[K[:, 0], 0])))
            ), 2), axis=1
    ))) / 2

    CH_AD = Surface / CHarea

    VDs = np.transpose(
        [MALs[0], MALs[1], MALs[2], Elong, Fltns, VDaabb, ADaabb, VDombb, ADombb, VDaee, ADaee, VDmvee, ADmvee,
         Solidity, CH_AD])

    return VDs, K


def getMax3Ddiam(ROIboxV, ConvHullV):
    n_v_int = ROIboxV.shape[0]

    if 1 <= n_v_int < 1000:
        ROIboxV = ROIboxV.astype(np.float32)
        NumVer = ROIboxV.shape[0]

        # Equivalent to tiling + reshaping with order='F'
        diff = ROIboxV[:, np.newaxis, :] - ROIboxV[np.newaxis, :, :]
        dist_sq = np.sum(diff ** 2, axis=2)
        Max3Ddiam = np.sqrt(np.max(dist_sq))

    elif n_v_int >= 1000:
        idx0 = ConvHullV[:, 0].astype(int)
        idx1 = ConvHullV[:, 1].astype(int)
        idx2 = ConvHullV[:, 2].astype(int)

        r1 = np.column_stack((
            ROIboxV[idx0, 0],
            ROIboxV[idx1, 1],
            ROIboxV[idx2, 2]
        )).astype(np.float32)

        NumVer = r1.shape[0]
        diff = r1[:, np.newaxis, :] - r1[np.newaxis, :, :]
        dist_sq = np.sum(diff ** 2, axis=2)
        Max3Ddiam = np.sqrt(np.max(dist_sq))

    else:
        Max3Ddiam = 0.0

    return Max3Ddiam


def geospatial(df_int, spacing, xI, yI, zI):
    n_v = df_int.shape[0]
    pos_mat = np.column_stack((zI * spacing[0], yI * spacing[1], xI * spacing[2]))

    gl_dev = df_int - np.mean(df_int)
    gl_denom = np.sum(gl_dev ** 2.0)

    if n_v < 2000:
        # Match original indexing logic using tile and repeat
        comb_iter = np.array([np.tile(np.arange(n_v), n_v), np.repeat(np.arange(n_v), n_v)])
        comb_iter = comb_iter[:, comb_iter[0, :] > comb_iter[1, :]]

        i = comb_iter[0]
        j = comb_iter[1]

        diffs = pos_mat[i] - pos_mat[j]
        dist = np.sqrt(np.sum(diffs ** 2, axis=1))
        w_ij = 1.0 / dist

        # Moran's I
        moran_nom = n_v * np.sum(w_ij * gl_dev[i] * gl_dev[j])
        w_sum = np.sum(w_ij)

        moran_i = moran_nom / (w_sum * gl_denom) if gl_denom > 0.0 else 1.0

        # Geary's C
        geary_nom = (n_v - 1.0) * np.sum(w_ij * (gl_dev[i] - gl_dev[j]) ** 2)
        geary_c = geary_nom / (2 * w_sum * gl_denom) if gl_denom > 0.0 else 1.0

    else:
        moran_nom = 0.0
        geary_nom = 0.0
        w_denom = 0.0

        for ii in range(n_v - 1):
            jj = np.arange(ii + 1, n_v)
            diffs = pos_mat[ii] - pos_mat[jj]
            w_iijj = 1.0 / np.sqrt(np.sum(diffs ** 2, axis=1))

            gl_ii = gl_dev[ii]
            gl_jj = gl_dev[jj]

            moran_nom += np.sum(w_iijj * gl_ii * gl_jj)
            geary_nom += np.sum(w_iijj * (gl_ii - gl_jj) ** 2)
            w_denom += np.sum(w_iijj)

        moran_i = n_v * moran_nom / (w_denom * gl_denom) if gl_denom > 0.0 else 1.0
        geary_c = (n_v - 1.0) * geary_nom / (2 * w_denom * gl_denom) if gl_denom > 0.0 else 1.0

    return moran_i, geary_c


def getAutoCorrs(ROIints, xI, yI, zI, pixelW, sliceS):
    n_v_int = ROIints.size
    if (1000 > n_v_int >= 1):

        MoranI, GearyC = geospatial(df_int=ROIints, spacing=[pixelW, pixelW, sliceS], xI=xI, yI=yI, zI=zI)

    elif n_v_int >= 1000:
        moran_list, geary_list = [], []
        iter_nr = 1
        tol_aim = 0.002
        tol_sem = 1.000
        while tol_sem > tol_aim:
            curr_points = np.random.choice(n_v_int, size=100, replace=False)
            moran_i, geary_c = geospatial(df_int=ROIints[curr_points], spacing=[pixelW, pixelW, sliceS], xI=xI, yI=yI,
                                          zI=zI)

            moran_list.append(moran_i)
            geary_list.append(geary_c)

            if iter_nr > 10:
                tol_sem = np.max([np.std(moran_list), np.std(geary_list)]) / np.sqrt(iter_nr)
            iter_nr += 1

            del curr_points, moran_i, geary_c

        MoranI = np.mean(moran_list)
        GearyC = np.mean(geary_list)

        del iter_nr

    return MoranI, GearyC





import numpy as np
import scipy.ndimage as ndi
import SimpleITK as sitk
import gc
from ...utils.utils import save_numpy_on_disk
from typing import Optional, Dict, Any, List, Tuple, Union
# -------------------------------------------------------------------------
# function [Vnew] = imresize3D(V,scale,tsize,ntype,npad,new_spacing,isIsot2D)
# -------------------------------------------------------------------------
# DESCRIPTION:
# This function resizes a 3D input volume to new dimensions.
# -------------------------------------------------------------------------
# INPUTS:
# - V: The input volume (3D array).
# - scale: scaling factor, when used set tsize to [].
# - nsize: new dimensions, when used set scale to [].
# - ntype: Type of interpolation ('nearest', 'linear', or 'cubic')
# - npad: Boundary condition ('replicate', 'symmetric', 'circular', 'fill',
#         or 'bound')
# -------------------------------------------------------------------------
# OUTPUTS:
# - Vnew: Resized volume.
# -------------------------------------------------------------------------
# AUTHOR(S): 
# - Mahdi Hosseinzadeh
# -------------------------------------------------------------------------
# HISTORY:
# - Creation: December 2022
#--------------------------------------------------------------------------



def imresize3D(input_V:Union[str, np.ndarray],scale,tsize,ntype,npad,new_spacing,isIsot2D, use_disk=True):

    # I = np.eye(4)
    # print(I)s
    # I[0,0] = scale[0]
    # I[1,1] = scale[1]
    # I[2,2] = scale[2]
    # print(I)
    # Vnew = imwarp(V,I,ntype)
    # Vnew = skimage.transform.warp(V,I,ntype)
    # Read order of multidimensional spline filter (0=nearest neighbours, 1=linear, 3=cubic)
    # Vnew = translate_image(V, I, ntype=ntype)
    # Load from disk
    if use_disk:
        V = np.load(input_V, mmap_mode='r')
    else:
        V = input_V

    new_spacing = np.array(new_spacing)
    # scale = np.array(scale , dtype=np.float32).flatten(order='F')
    new_spacing = new_spacing.astype(np.float32)
    # scale = scale.astype(np.float32)
    ntype = ntype.lower()
    if ntype == 'nearest':
        order = 0
    elif ntype == 'linear':
        order = 1
    elif ntype == 'cubic':
        order = 3
    else:
        order = 0


    
    if isIsot2D == 1:
        new_spacing[2] = scale[2]

    Vnew = interpolate_to_new_grid(         # toto returns path
                            orig_dim=np.asarray(V.shape),
                            orig_spacing=np.asarray(scale).flatten(order='F'),
                            orig_vox_path=input_V,
                            grid_origin = None,
                            sample_spacing = np.asarray(new_spacing).flatten(order='F'),
                            order=order,
                            mode=npad,
                            align_to_center=True,
                            use_disk=use_disk
                            )

    # Clean RAM
    del V, input_V
    gc.collect()

    # orientation = (1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0)
    # m_affine = np.zeros((3, 3), dtype=float)
    # m_affine[:, 0] = scale[0] * np.array([orientation[0], orientation[1], orientation[2]])
    # m_affine[:, 1] = scale[1] * np.array([orientation[3], orientation[4], orientation[5]])
    # m_affine[:, 2] = scale[2] * np.array([orientation[6], orientation[7], orientation[8]])
    # m_affine_inv = np.linalg.inv(m_affine)


    # interpolate_to_new_grid(
    #                     # sample_dim=img_obj.size,
    #                     grid_origin=np.dot(m_affine_inv, np.transpose(img_obj.origin - self.roi.origin)),
    #                     align_to_center=False
    #                     )

    # Vnew = interpolate(
    #                 vol_dim=np.asarray(V.shape),
    #                 vol_spacing=np.asarray(scale).flatten(),
    #                 vol=V,
    #                 grid_origin = None,
    #                 new_dim = np.asarray(tsize).flatten(),
    #                 new_spacing = np.asarray(new_spacing).flatten(),
    #                 order=order,
    #                 npad=npad)
    
    

    return Vnew

# -------------------------------------------------------------------------
# function [Vnew] = imresize(V,scale,tsize,ntype,npad,new_spacing,isIsot2D)
# -------------------------------------------------------------------------
# DESCRIPTION:
# This function resizes a 2D input volume to new dimensions.
# -------------------------------------------------------------------------
# INPUTS:
# - V: The input volume (2D array).
# - scale: scaling factor, when used set tsize to [].
# - nsize: new dimensions, when used set scale to [].
# - ntype: Type of interpolation ('nearest', 'linear', or 'cubic')
# - npad: Boundary condition ('replicate', 'symmetric', 'circular', 'fill',
#         or 'bound')
# -------------------------------------------------------------------------
# OUTPUTS:
# - Vnew: Resized volume.
# -------------------------------------------------------------------------
# AUTHOR(S): 
# - Mahdi Hosseinzadeh
# -------------------------------------------------------------------------
# HISTORY:
# - Creation: December 2022
#--------------------------------------------------------------------------

def imresize(V_path, scale, tsize, ntype, new_spacing, use_disk=True):
    # Load from disk
    if use_disk:
        V = np.load(V_path, mmap_mode='r')
    else:
        V = V_path

    new_spacing[2] = scale[2]
    new_spacing = np.array(new_spacing)
    # scale = np.array(scale,dtype=np.float32)
    new_spacing = new_spacing.astype(np.float32)
    # scale = scale.astype(np.float32)

    ntype = ntype.lower()
    if ntype == 'nearest':
        order = 0
    elif ntype == 'linear':
        order = 1
    elif ntype == 'cubic':
        order = 2
    else:
        order = 0
    

    Vnew = interpolate_to_new_grid(
                            orig_dim=np.asarray(V.shape),
                            orig_spacing=np.asarray(scale).flatten(order='F'),
                            orig_vox_path=V_path,
                            grid_origin = None,
                            sample_spacing = np.asarray(new_spacing).flatten(order='F'),
                            order=order,
                            use_disk=use_disk)

    # Clean RAM
    del V, V_path
    gc.collect()
    # Vnew = interpolate(
    #                 vol_dim=np.asarray(V.shape),
    #                 vol_spacing=np.asarray(scale).flatten(),
    #                 vol=V,
    #                 new_dim = np.asarray(tsize).flatten(),
    #                 new_spacing = new_spacing,
    #                 order=order)

    return Vnew


def interpolate_to_new_grid(orig_dim,
                            orig_spacing,
                            orig_vox_path:str,
                            sample_dim = None ,
                            sample_spacing = np.array([0.0, 0.0, 0.0]),
                            grid_origin = None,
                            translation = np.array([0.0, 0.0, 0.0]),
                            order = 0,
                            mode='nearest',
                            align_to_center=True,
                            processor='scipy',
                            use_disk=True):

    # Load from disk
    if use_disk:
        orig_vox = np.load(orig_vox_path, mmap_mode='r')
    else:
        orig_vox = orig_vox_path
        # Clean RAM
        del orig_vox_path
        gc.collect()

    sample_spacing = sample_spacing.astype(np.float32)
    orig_spacing = orig_spacing.astype(np.float32)
    orig_dim = orig_dim.astype(np.float32)

    if sample_dim is None:
        sample_dim = np.ceil(np.multiply(orig_dim, orig_spacing / sample_spacing))
    #     sample_dim = np.array(sample_dim[0])

    grid_spacing = sample_spacing / orig_spacing

    if grid_origin is None:
        if align_to_center:
            grid_origin = 0.5 * (np.array(orig_dim) - 1.0) - 0.5 * (np.array(sample_dim) - 1.0) * grid_spacing
        else:
            grid_origin = np.array([0.0, 0.0, 0.0])

        grid_origin += translation * grid_spacing


    if processor == 'scipy':
        map_x, map_y, map_z = np.mgrid[:sample_dim[0], :sample_dim[1], :sample_dim[2]]

        map_z = map_z * grid_spacing[2] + grid_origin[2]
        map_z = map_z.astype(np.float32)
        map_y = map_y * grid_spacing[1] + grid_origin[1]
        map_y = map_y.astype(np.float32)
        map_x = map_x * grid_spacing[0] + grid_origin[0]
        map_x = map_x.astype(np.float32)

        map_vox = ndi.map_coordinates(input=orig_vox.astype(np.float32),
                                        coordinates=np.array([map_x, map_y, map_z], dtype=np.float32),
                                        order=order,
                                        mode=mode)

    elif processor == 'sitk':
        sitk_orig_img = sitk.GetImageFromArray(orig_vox.astype(np.float32), isVector=False)
        sitk_orig_img.SetOrigin(np.array([0.0, 0.0, 0.0]))
        sitk_orig_img.SetSpacing(np.array(orig_spacing,dtype=np.float32).tolist())

        interpolator = sitk.ResampleImageFilter()

        if order == 0:
            interpolator.SetInterpolator(sitk.sitkNearestNeighbor)
        elif order == 1:
            interpolator.SetInterpolator(sitk.sitkLinear)
        elif order == 2:
            interpolator.SetInterpolator(sitk.sitkBSplineResamplerOrder2)
        elif order == 3:
            interpolator.SetInterpolator(sitk.sitkBSpline)


        interpolator.SetOutputOrigin(np.array(grid_origin[::-1],dtype=np.float32).tolist())
        interpolator.SetOutputSpacing(np.array(grid_spacing[::-1],dtype=np.float32).tolist())
        interpolator.SetSize(sample_dim[::-1].astype(int).tolist())

        map_vox = sitk.GetArrayFromImage(interpolator.Execute(sitk_orig_img))
    # Save array on disk
    if use_disk:
        map_vox_path = save_numpy_on_disk(map_vox, prefix='map_vox', suffix='.npy')
    else:
        map_vox_path = map_vox

    # Clean RAM
    del orig_vox, map_vox
    gc.collect()
    return map_vox_path





import numpy as np



# -------------------------------------------------------------------------
# function [textures] = getGlobalTextures(ROIonly,Nbins)
# -------------------------------------------------------------------------
# DESCRIPTION: 
# This function computes Global texture features from the region of 
# interest (ROI) of an input volume.
# -------------------------------------------------------------------------
# INPUTS:
# - ROIonly: The smallest box containing the resampled 3D ROI, with the
#            imaging data ready for texture analysis computations. Voxels
#            outside the ROI are set to NaNs.
# - BinSize: number of bins (for fixed number of bins method) or bin size
#            (for bin size method) 
# - DiscType: Disctritization type. Either 'FNB' for fixed number of bins
#             or 'FBS' for fixed bin size. 
#
# -------------------------------------------------------------------------
# OUTPUTS:
# - HistVect: Vector specifying the values of different Intensity Histogram 
#             features as defined below.
# -------------------------------------------------------------------------
# AUTHOR(S): 
# - Saeed Ashrafinia
# - Mahdi Hosseinzadeh
# -------------------------------------------------------------------------
# HISTORY:
# - Creation: December 2022
# -------------------------------------------------------------------------

def getHist(ROIonly2,BinSize, DiscType, feature_value_mode='REAL_VALUE', epsilon=1e-30):
    
    ROIonly = ROIonly2.copy()
    ROIarrayValid = ROIonly[~np.isnan(ROIonly2)].flatten(order='F')

    if DiscType == 'FBN':
        histraw = np.histogram(ROIarrayValid,BinSize)
        histo = np.divide(histraw[0],np.sum(histraw[0]))
        Binslist = np.arange(1,BinSize+1)
        Mean = np.sum(histo * Binslist)

    elif DiscType == 'FBS':
        
        BinsCntrVect = np.arange(1,np.max(ROIarrayValid)+2)
        histraw = np.histogram(ROIarrayValid ,BinsCntrVect)
        histo = np.divide(histraw[0],np.sum(histraw[0]))
        BinsCntrVect = np.delete(BinsCntrVect,-1)
        # BinSize = BinsCntrVect.ndim
        BinSize = BinsCntrVect.shape[0]
        Binslist = BinsCntrVect
        Mean = np.sum(histo * Binslist)


    Mean = Mean

    Var = 0
    skw = 0
    krt = 0
    entropy = 0
    for i in range(0,BinSize):
        Var = Var + histo[i] *  np.float_power(((i+1)-Mean),2)
        skw = skw + histo[i] *  np.float_power(((i+1)-Mean),3)
        krt = krt + histo[i] *  np.float_power(((i+1)-Mean),4)
        entropy = entropy-histo[i] *   np.log2(   histo[i] + epsilon)
    
    Entropy = entropy

    sigma = np.sqrt(Var)
    Variance = Var

    skw = skw / np.float_power(sigma,3)
    if np.isnan(skw) == True:
        skw = 0
    Skewness = skw
    krt = (krt   /  np.float_power(sigma,4)) - 3

    if np.isnan(krt) == True:
        krt = 0
    Kurtosis = krt

    SUVmax = np.max(Binslist)

    cumHisto = 0
    Prcnt10 = float('inf')
    Prcnt90 = float('inf')
    Median = float('inf')
    q75 = float('inf')
    q25 = float('inf')
    for i in range(0,BinSize):
        cumHisto += histo[i]
        if cumHisto >= 0.9:
            if (i+1) <= Prcnt90:
                Prcnt90 = i+1
        if cumHisto >= 0.75:
            if (i+1) <= q75:
                q75 = i+1
        if cumHisto >= 0.5:
            if (i+1) <= Median:
                Median = i+1
        if cumHisto >= 0.25:
            if (i+1) <= q25:
                q25 = i+1
        if cumHisto >= 0.1:
            if (i+1) <= Prcnt10:
                Prcnt10 = i+1

    # SUVmin = np.min(Binslist)
    SUVmin = np.min(np.where(histraw[0]>0)[0])+1


    ModeIdx = np.array(np.argmax(histo))
    if ModeIdx.size == 1:
        Mode = ModeIdx + 1
    else:
        Idx = np.argsort(np.abs(ModeIdx-Mean))
        Mode = int(ModeIdx[np.argmin(Idx)]) + 1

    IqntlRange = q75-q25
    
    Range = SUVmax - SUVmin

    nV = ROIarrayValid.shape[0]
    MAD = np.sum(np.abs(    np.multiply( histraw[0],(Binslist - Mean) )    )) / nV

    RobustSet = list(np.arange(Prcnt10-1,Prcnt90))
    nVr = np.sum(histraw[0][RobustSet])

    aaa = np.divide(histraw[0][RobustSet] , nVr)
    bbb = np.transpose(Binslist[RobustSet])
    SUVmeanR =  np.dot (aaa,bbb)
    RMAD = np.sum(np.abs(    np.multiply(histraw[0][RobustSet],(Binslist[RobustSet] - SUVmeanR) )  )) / nVr


    MedAD = np.sum(np.abs(    np.multiply( histraw[0],(Binslist - Median) )   ))  / nV


    CoV = np.sqrt(Var) / Mean


    QCoD = (q75 - q25) / (q75 + q25)

    Energy = np.sum(  np.float_power(histo,2))

    
    
    try:
        HH = []
        H = histraw[0][1]-histraw[0][0]
        HH.append(H)
        H2 = list(np.divide( np.subtract(histraw[0][2:] , histraw[0][0:-2] ) , 2))
        HH = HH + H2
        H3 = histraw[0][-1] - histraw[0][-2] 
        HH.append(H3)
        HH = np.array(HH)
    except:
        HH = histraw[0][0]

    MaxGrad, MaxGradGL = np.max(HH),np.argmax(HH)+1

    MinGrad, MinGradGL = np.min(HH),np.argmin(HH)+1



    HistVect = [Mean, Variance, Skewness, Kurtosis, Median,
            SUVmin, Prcnt10, Prcnt90, SUVmax, Mode, 
            IqntlRange, Range, MAD, RMAD, MedAD, 
            CoV, QCoD, Entropy, Energy, 
            MaxGrad, MaxGradGL, MinGrad, MinGradGL]

    return HistVect


"""
pysera - A Python library for radiomics feature extraction with multiprocessing support

This library provides a simple, focused interface for radiomics feature extraction
from medical imaging data with built-in multiprocessing and logging capabilities.

Main Features:
- Support for multiple medical image formats (NIfTI, DICOM, NRRD, NumPy arrays)
- Built-in multiprocessing for efficient batch processing
- Comprehensive logging with Excel export functionality
- Extensive radiomics feature extraction capabilities
- Simple, single-function API

Example Usage:
    >>> import pysera
    >>> 
    >>> # Process single image-mask pair
    >>> result = pysera.process_batch(
    ...     image_input="image.nii.gz",
    ...     mask_input="mask.nii.gz",
    ...     output_path="./results"
    ... )
    >>> 
    >>> # Process batch with multiprocessing
    >>> result = pysera.process_batch(
    ...     image_input="./images",
    ...     mask_input="./masks",
    ...     output_path="./results",
    ...     num_workers="4",
    ...     feats2out=2,
    ...     apply_preprocessing=True
    ... )
"""

__version__ = "1.0.2"
__author__ = "Mohammad R. Salmanpour, Amir Hossein Pouria"
__email__ = "m.salmanpoor66@gmail.com"

# Import internal components (not exposed to users)
from .processing.radiomics_processor import RadiomicsProcessor
from .utils.log_logging import MemoryLogHandler
from .config.settings import LOGGING_CONFIG

def process_batch(
    image_input,
    mask_input,
    output_path=None,
    apply_preprocessing=False,
    min_roi_volume=10,
    num_workers="auto",
    enable_parallelism=True,
    feats2out=2,
    bin_size=25,
    roi_num=10,
    roi_selection_mode="per_Img",
    feature_value_mode="REAL_VALUE",
    report=True,
    optional_params=None
):
    """
    Process radiomics feature extraction for single or batch of medical images.
    
    This is the main function for pysera library that handles all radiomics processing
    with multiprocessing and logging capabilities built-in.
    
    Args:
        image_input (str): Path to image file or directory containing images.
                         Supported formats: NIfTI (.nii, .nii.gz), DICOM (.dcm), 
                         NRRD (.nrrd), NumPy (.npy), and directories with multiple files.
        
        mask_input (str): Path to mask file or directory containing masks.
                        Must have same format and dimensions as corresponding images.
        
        output_path (str, optional): Output directory for results. 
                                   Defaults to "./output_optimized" if not specified.
        
        apply_preprocessing (bool, optional): Apply ROI and intensity preprocessing.
                                            Defaults to False.
        
        min_roi_volume (int, optional): Minimum ROI volume threshold for processing.
                                      ROIs smaller than this will be filtered out.
                                      Defaults to 10.
        
        num_workers (str, optional): Number of parallel workers for multiprocessing.
                                   If auto, uses all available CPU cores.
                                   Set to 1 to disable multiprocessing.
                                   Defaults to None (auto-detect).
        
        enable_parallelism (bool, optional): Explicitly enable parallel processing.
                                         Defaults to True.
        
        feats2out (int, optional): Feature extraction mode (1-12).
                                 1: all IBSI_Evaluation, 2: 1st+3D+2.5D (default), 3: 1st+2D+2.5D,
                                 4: 1st+3D+selected2D+2.5D, 5: all+Moment, 6: 1st+2D+2.5D,
                                 7: 1st+2.5D, 8: 1st only, 9: 2D only, 10: 2.5D only,
                                 11: 3D only, 12: Moment only.
                                 Defaults to 2.
        
        bin_size (int, optional): Intensity discretization bin size for texture analysis.
                                Defaults to 25.
        
        roi_num (int, optional): Number of ROIs to select for processing.
                               Defaults to 10.
        
        roi_selection_mode (str, optional): ROI selection strategy.
                                          "per_Img": Select largest ROIs across entire image.
                                          "per_region": Group ROIs by color/label first.
                                          Defaults to "per_Img".
        
        feature_value_mode (str, optional): Value type for feature extraction.
                                  "APPROXIMATE_VALUE": Fast computation (recommended).
                                  "REAL_VALUE": Precise computation (slower).
                                  Defaults to "REAL_VALUE".
        
        report (bool, optional): Enable comprehensive logging during processing.
                                       Defaults to True.

    
    Returns:
        dict: Processing results containing:
            - 'success': bool indicating if processing completed successfully
            - 'output_path': str path to the output directory
            - 'processed_files': int number of files processed
            - 'features_extracted': int number of features extracted
            - 'processing_time': float total processing time in seconds
            - 'logs': list of log messages if logging was enabled
    
    Example:
        >>> import pysera
        >>> 
        >>> # Basic usage
        >>> result = pysera.process_batch(
        ...     image_input="scan.nii.gz",
        ...     mask_input="mask.nii.gz",
        ...     output_path="./results"
        ... )
        >>> 
        >>> # Advanced usage with multiprocessing and custom parameters
        >>> result = pysera.process_batch(
        ...     image_input="./patient_scans",
        ...     mask_input="./patient_masks",
        ...     output_path="./radiomics_results",
        ...     num_workers="4",
        ...     feats2out=5,  # Extract all features + Moment
        ...     min_roi_volume=10,
        ...     apply_preprocessing=True,
        ... )
        >>> 
        >>> print(f"Success: {result['success']}")
        >>> print(f"Processed {result['processed_files']} files")
        >>> print(f"Results saved to: {result['output_path']}")
    
    Raises:
        FileNotFoundError: If image_input or mask_input do not exist.
        ValueError: If parameters are invalid (e.g., feats2out not in range 1-12).
        RuntimeError: If processing fails due to incompatible image/mask dimensions.
    
    Note:
        This function automatically handles:
        - File format detection and loading
        - Multiprocessing setup and management
        - Logging configuration and Excel export
        - ROI preprocessing and optimization
        - Feature extraction and result compilation
        - Error handling and cleanup
    """
    import time
    from pathlib import Path
    import logging
    logger = logging.getLogger("Dev_logger")

    start_time = time.time()
    
    # Set up logging if requested
    memory_handler = None
    if report:
        memory_handler = MemoryLogHandler()
        logger = _setup_logging(memory_handler=memory_handler)
        logger.info("Starting pysera radiomics processing")
    
    try:
        # Create processor with logging if enabled
        processor = RadiomicsProcessor(
            output_path=output_path,
            memory_handler=memory_handler if report else None
        )
        
        # Process the batch
        result = processor.process_batch(
            image_input=image_input,
            mask_input=mask_input,
            apply_preprocessing=apply_preprocessing,
            min_roi_volume=min_roi_volume,
            num_workers=num_workers,
            enable_parallelism=enable_parallelism,
            feats2out=feats2out,
            bin_size=bin_size,
            roi_num=roi_num,
            roi_selection_mode=roi_selection_mode,
            feature_value_mode=feature_value_mode,
            optional_params=optional_params
        )
        
        processing_time = time.time() - start_time
        
        # Prepare return results
        return_result = {
            'success': bool(result),
            'output_path': processor.output_path,
            'processed_files': result.get('processed_files', 0) if result else 0,
            'features_extracted': result["out"][1] if result else 0,
            'processing_time': processing_time,
            'logs': memory_handler.get_logs() if memory_handler else []
        }
        
        if report:
            if result:
                logger.info(f"Processing completed successfully in {processing_time:.2f} seconds")
            else:
                logger.error("Processing failed")
        
        return return_result
        
    except Exception as e:
        processing_time = time.time() - start_time
        
        if report and 'logger' in locals():
            logger.error(f"Processing failed with error: {e}")
        
        return {
            'success': False,
            'output_path': output_path,
            'processed_files': 0,
            'features_extracted': 0,
            'processing_time': processing_time,
            'logs': memory_handler.get_logs() if memory_handler else [],
            'error': str(e)
        }


def _setup_logging(memory_handler=None):
    """Internal function to set up logging configuration."""
    import logging
    
    logger = logging.getLogger('Dev_logger')
    logger.setLevel(logging.DEBUG)

    if logger.hasHandlers():
        logger.handlers.clear()

    # Console handler
    console_handler = logging.StreamHandler()
    console_handler.setLevel(getattr(logging, LOGGING_CONFIG['console_level']))
    console_format = logging.Formatter(LOGGING_CONFIG['console_format'])
    console_handler.setFormatter(console_format)
    logger.addHandler(console_handler)

    # Memory handler (write in excel)
    if memory_handler:
        memory_handler.setLevel(getattr(logging, LOGGING_CONFIG['memory_level']))
        memory_format = logging.Formatter(LOGGING_CONFIG['console_format'])
        memory_handler.setFormatter(memory_format)
        logger.addHandler(memory_handler)

    return logger


# Feature extraction modes information (for documentation/reference)
_FEATURE_EXTRACTION_MODES = {
    1: "all IBSI_Evaluation",
    2: "1st+3D+2.5D (default)", 
    3: "1st+2D+2.5D",
    4: "1st+3D+selected2D+2.5D",
    5: "all+Moment",
    6: "1st+2D+2.5D",
    7: "1st+2.5D",
    8: "1st only",
    9: "2D only",
    10: "2.5D only",
    11: "3D only",
    12: "Moment only"
}

# Main exports - only process_batch is public
__all__ = [
    'process_batch',  # Main public function
    '__version__',
    '__author__',
    '__email__'
] 